# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json
from typing import Optional, List

from django.conf import settings
from django.db import transaction

from kombu import Exchange, Queue, Connection

from bamboo_engine import states
from bamboo_engine.template import Template
from bamboo_engine.eri import EngineRuntimeInterface, NodeType, ContextValueType

from pipeline.eri import codec
from pipeline.eri.imp.plugin_manager import PipelinePluginManagerMixin
from pipeline.eri.imp.hooks import HooksMixin
from pipeline.eri.imp.process import ProcessMixin
from pipeline.eri.imp.node import NodeMixin
from pipeline.eri.imp.state import StateMixin
from pipeline.eri.imp.schedule import ScheduleMixin
from pipeline.eri.imp.data import DataMixin
from pipeline.eri.imp.context import ContextMixin
from pipeline.eri.imp.execution_history import ExecutionHistoryMixin
from pipeline.eri.imp.task import TaskMixin
from pipeline.eri.celery.queues import QueueResolver

from pipeline.eri.models import Node, Data, ContextValue, Process, ContextOutputs, LogEntry, ExecutionHistory, State


class BambooDjangoRuntime(
    TaskMixin,
    ExecutionHistoryMixin,
    ContextMixin,
    DataMixin,
    ScheduleMixin,
    StateMixin,
    NodeMixin,
    ProcessMixin,
    PipelinePluginManagerMixin,
    HooksMixin,
    EngineRuntimeInterface,
):
    CONTEXT_VALUE_TYPE_MAP = {
        "plain": ContextValueType.PLAIN.value,
        "splice": ContextValueType.SPLICE.value,
        "lazy": ContextValueType.COMPUTE.value,
    }

    def _data_inputs_assemble(self, pipeline_id: str, node_id: str, node_inputs: dict) -> (dict, List[ContextValue]):
        inputs = {}
        context_values = []
        for k, v in node_inputs.items():
            if v["type"] == "lazy":
                cv_key = "${%s_%s}" % (k, node_id)
                if len(cv_key) > 128:
                    raise ValueError("var key %s length exceeds 128" % cv_key)
                context_values.append(
                    ContextValue(
                        pipeline_id=pipeline_id,
                        key=cv_key,
                        type=ContextValueType.COMPUTE.value,
                        serializer=self.JSON_SERIALIZER,
                        value=json.dumps(v["value"]),
                        code=v.get("custom_type", ""),
                    )
                )
                inputs[k] = {"need_render": True, "value": cv_key}
            else:
                inputs[k] = {"need_render": v["type"] == "splice", "value": v["value"]}

        return inputs, context_values

    def _gen_executable_end_event_node(self, event: dict, pipeline: dict, root_id: str, parent_id: str) -> Node:
        return Node(
            node_id=event["id"],
            detail=json.dumps(
                {
                    "id": event["id"],
                    "type": NodeType.ExecutableEndEvent.value,
                    "targets": {},
                    "root_pipeline_id": root_id,
                    "parent_pipeline_id": parent_id,
                    "can_skip": False,
                    "can_retry": True,
                    "code": event["type"],
                }
            ),
        )

    def _gen_event_node(self, event: dict, pipeline: dict, root_id: str, parent_id: str) -> Node:
        return Node(
            node_id=event["id"],
            detail=json.dumps(
                {
                    "id": event["id"],
                    "type": event["type"],
                    "targets": {event["outgoing"]: pipeline["flows"][event["outgoing"]]["target"]}
                    if event["type"] == NodeType.EmptyStartEvent.value
                    else {},
                    "root_pipeline_id": root_id,
                    "parent_pipeline_id": parent_id,
                    "can_skip": event["type"] == NodeType.EmptyStartEvent.value,
                    "can_retry": True,
                }
            ),
        )

    def _gen_gateway_node(self, gateway: dict, pipeline: dict, root_id: str, parent_id: str) -> Node:
        if gateway["type"] != NodeType.ConvergeGateway.value:
            targets = {flow_id: pipeline["flows"][flow_id]["target"] for flow_id in gateway["outgoing"]}
        else:
            targets = {gateway["outgoing"]: pipeline["flows"][gateway["outgoing"]]["target"]}

        detail = {
            "id": gateway["id"],
            "type": gateway["type"],
            "targets": targets,
            "root_pipeline_id": root_id,
            "parent_pipeline_id": parent_id,
            "can_retry": True,
            "can_skip": False,
        }

        if gateway["type"] == NodeType.ExclusiveGateway.value:
            detail["can_skip"] = True
            detail["conditions"] = [
                {
                    "name": flow_id,
                    "evaluation": cond["evaluate"],
                    "target_id": pipeline["flows"][flow_id]["target"],
                    "flow_id": flow_id,
                }
                for flow_id, cond in gateway["conditions"].items()
            ]
        elif gateway["type"] == NodeType.ParallelGateway.value:
            detail["converge_gateway_id"] = gateway["converge_gateway_id"]

        elif gateway["type"] == NodeType.ConditionalParallelGateway.value:
            detail["conditions"] = [
                {
                    "name": flow_id,
                    "evaluation": cond["evaluate"],
                    "target_id": pipeline["flows"][flow_id]["target"],
                    "flow_id": flow_id,
                }
                for flow_id, cond in gateway["conditions"].items()
            ]
            detail["converge_gateway_id"] = gateway["converge_gateway_id"]
        elif gateway["type"] == NodeType.ConvergeGateway.value:
            pass
        else:
            raise ValueError("unsupport gateway type {}: {}".format(gateway["type"], gateway))

        return Node(node_id=gateway["id"], detail=json.dumps(detail))

    def _gen_activity_node(self, act: dict, pipeline: dict, root_id: str, parent_id: str) -> Node:
        return Node(
            node_id=act["id"],
            detail=json.dumps(
                {
                    "id": act["id"],
                    "type": NodeType.ServiceActivity.value,
                    "targets": {act["outgoing"]: pipeline["flows"][act["outgoing"]]["target"]},
                    "root_pipeline_id": root_id,
                    "parent_pipeline_id": parent_id,
                    "can_skip": act["skippable"],
                    "code": act["component"]["code"],
                    "version": act["component"].get("version", "legacy"),
                    "timeout": act.get("timeout"),
                    "error_ignorable": act["error_ignorable"],
                    "can_retry": act["retryable"],
                }
            ),
        )

    def _gen_subproc_node(self, subproc: dict, pipeline: dict, root_id: str, parent_id: str) -> Node:
        return Node(
            node_id=subproc["id"],
            detail=json.dumps(
                {
                    "id": subproc["id"],
                    "type": NodeType.SubProcess.value,
                    "targets": {subproc["outgoing"]: pipeline["flows"][subproc["outgoing"]]["target"]},
                    "root_pipeline_id": root_id,
                    "parent_pipeline_id": parent_id,
                    "can_skip": False,
                    "can_retry": True,
                    "start_event_id": subproc["pipeline"]["start_event"]["id"],
                }
            ),
        )

    def _prepare(
        self, pipeline: dict, root_id: str, parent_id: Optional[str] = None
    ) -> (List[Node], List[Data], List[ContextValue], List[ContextOutputs]):

        parent_id = parent_id or root_id

        nodes = []
        datas = []
        context_values = []
        context_outputs = []

        node_outputs = {}
        context_var_references = {}
        final_references = {}

        # collect all node outputs and initial reference
        for key, input_data in pipeline["data"]["inputs"].items():
            source_act = input_data.get("source_act")
            source_key = input_data.get("source_key")
            if not source_act:
                context_var_references[key] = Template(input_data["value"]).get_reference()
                final_references[key] = set()
                context_values.append(
                    ContextValue(
                        pipeline_id=pipeline["id"],
                        key=key,
                        type=self.CONTEXT_VALUE_TYPE_MAP[input_data["type"]],
                        serializer=self.JSON_SERIALIZER,
                        value=json.dumps(input_data["value"]),
                        code=input_data.get("custom_type", ""),
                    )
                )
            else:
                if isinstance(source_act, list):
                    for sa in source_act:
                        node_outputs.setdefault(sa["source_act"], {})[sa["source_key"]] = key
                else:
                    node_outputs.setdefault(source_act, {})[source_key] = key

        # process activities
        for act in pipeline["activities"].values():
            if act["type"] == NodeType.ServiceActivity.value:
                # node
                nodes.append(self._gen_activity_node(act=act, pipeline=pipeline, root_id=root_id, parent_id=parent_id))
                # data
                data_inputs, compute_cvs = self._data_inputs_assemble(parent_id, act["id"], act["component"]["inputs"])
                datas.append(
                    Data(
                        node_id=act["id"],
                        inputs=codec.data_json_dumps(data_inputs),
                        outputs=json.dumps(node_outputs.get(act["id"], {})),
                    )
                )
                # compute context values
                for cv in compute_cvs:
                    context_values.append(cv)
                    final_references[cv.key] = set()
                    context_var_references[cv.key] = Template(cv.value).get_reference()

            elif act["type"] == NodeType.SubProcess.value:
                # node
                nodes.append(
                    self._gen_subproc_node(subproc=act, pipeline=pipeline, root_id=root_id, parent_id=parent_id)
                )
                # data
                data_inputs, compute_cvs = self._data_inputs_assemble(parent_id, act["id"], act["params"])
                datas.append(
                    Data(
                        node_id=act["id"],
                        inputs=codec.data_json_dumps(data_inputs),
                        outputs=json.dumps(node_outputs.get(act["id"], {})),
                    )
                )
                # compute context values
                for cv in compute_cvs:
                    context_values.append(cv)
                    final_references[cv.key] = set()
                    context_var_references[cv.key] = Template(cv.value).get_reference()

                # subprocess output
                context_outputs.append(
                    ContextOutputs(pipeline_id=act["id"], outputs=json.dumps(act["pipeline"]["data"]["outputs"]))
                )
                sub_nodes, sub_datas, sub_ctx_values, sub_ctx_outputs = self._prepare(
                    pipeline=act["pipeline"], root_id=root_id, parent_id=act["id"]
                )

                nodes.extend(sub_nodes)
                datas.extend(sub_datas)
                context_values.extend(sub_ctx_values)
                context_outputs.extend(sub_ctx_outputs)
            else:
                raise ValueError("unsupport act type {}: {}".format(act["type"], act["id"]))

        # process events
        nodes.append(
            self._gen_event_node(event=pipeline["start_event"], pipeline=pipeline, root_id=root_id, parent_id=parent_id)
        )
        if pipeline["end_event"]["type"] == NodeType.EmptyEndEvent.value:
            nodes.append(
                self._gen_event_node(
                    event=pipeline["end_event"], pipeline=pipeline, root_id=root_id, parent_id=parent_id
                )
            )
        else:
            nodes.append(
                self._gen_executable_end_event_node(
                    event=pipeline["end_event"], pipeline=pipeline, root_id=root_id, parent_id=parent_id
                )
            )

        # process gateways
        for gateway in pipeline["gateways"].values():
            nodes.append(
                self._gen_gateway_node(gateway=gateway, pipeline=pipeline, root_id=root_id, parent_id=parent_id)
            )

        # resolve final references (BFS)
        # convert a:b, b:c,d -> a:b,c,d b:c,d
        for key, references in context_var_references.items():
            queue = []
            queue.extend(references)

            while queue:
                r = queue.pop()

                # processed
                if r in final_references[key]:
                    continue

                final_references[key].add(r)
                if r in context_var_references:
                    queue.extend(context_var_references[r])

        for cv in context_values:
            if cv.pipeline_id != parent_id:
                continue
            fr = final_references.get(cv.key)
            cv.references = json.dumps(list(fr)) if fr else "[]"

        if parent_id == root_id:
            context_outputs.append(ContextOutputs(pipeline_id=root_id, outputs=json.dumps(pipeline["data"]["outputs"])))

        return nodes, datas, context_values, context_outputs

    def prepare_run_pipeline(
        self, pipeline: dict, root_pipeline_data: dict, root_pipeline_context: dict, **options
    ) -> int:
        """
        进行 pipeline 执行前的准备工作，并返回 进程 ID，该函数执行完成后即代表
        pipeline 是随时可以通过 execute(process_id, start_event_id) 启动执行的
        一般来说，应该完成以下工作:
        - 准备好进程模型
        - 准备好流程中每个节点的信息
        - 准备好流程中每个节点数据对象的信息

        :param pipeline: pipeline 描述对象
        :type pipeline: dict
        :param root_pipeline_data 根流程数据
        :type root_pipeline_data: dict
        :param root_pipeline_context 根流程上下文
        :type root_pipeline_context: dict
        :return: 进程 ID
        :rtype: str
        """

        queue = options.get("queue", "")
        priority = options.get("priority", 100)
        pipeline_id = pipeline["id"]

        nodes, datas, context_values, context_outputs = self._prepare(pipeline=pipeline, root_id=pipeline["id"])
        datas.append(
            Data(
                node_id=pipeline_id,
                inputs=codec.data_json_dumps(
                    {k: {"need_render": False, "value": v} for k, v in root_pipeline_data.items()}
                ),
                outputs="{}",
            )
        )
        for key, value in root_pipeline_context.items():
            serialized, serializer = self._serialize(value)
            context_values.append(
                ContextValue(
                    pipeline_id=pipeline_id,
                    key=key,
                    type=self.CONTEXT_VALUE_TYPE_MAP["plain"],
                    serializer=serializer,
                    value=serialized,
                    references="[]",
                )
            )
        batch_size = getattr(settings, "BAMBOO_DJANGO_ERI_PREPARE_BATCH_SIZE", 500)

        with transaction.atomic():
            pid = Process.objects.create(
                root_pipeline_id=pipeline_id,
                queue=queue,
                priority=priority,
                pipeline_stack='["{}"]'.format(pipeline_id),
            ).id
            self.set_state(
                node_id=pipeline_id, to_state=states.RUNNING, root_id=pipeline_id, parent_id="", set_started_time=True,
            )

            Node.objects.bulk_create(nodes, batch_size=batch_size)
            Data.objects.bulk_create(datas, batch_size=batch_size)
            ContextValue.objects.bulk_create(context_values, batch_size=batch_size)
            ContextOutputs.objects.bulk_create(context_outputs, batch_size=batch_size)

        return pid

    def node_rerun_limit(self, root_pipeline_id: str, node_id: str) -> int:
        """
        返回节点最大重入次数

        :param root_pipeline_id: 根流程 ID
        :type root_pipeline_id: str
        :param node_id: 节点 ID
        :type node_id: str
        :return: 节点最大重入次数
        :rtype: int
        """
        return int(getattr(settings, "BAMBOO_DJANGO_ERI_NODE_RERUN_LIMIT", 100))

    def add_queue(self, name: str, routing_key: Optional[str] = ""):
        """
        在 Broker 中新增用户自定义队列，注意配合 CELERY_CREATE_MISSING_QUEUES 选项使用

        :param name: 队列名
        :type name: str
        :param routing_key: routing key
        :type routing_key: str
        """
        queue_resolver = QueueResolver(name)

        exchange = Exchange("default", type="direct")
        with Connection(settings.BROKER_URL) as conn:
            with conn.channel() as channel:
                for queue_config in queue_resolver.routes_config().values():
                    queue = Queue(
                        queue_config["queue"], exchange, routing_key=queue_config["routing_key"], max_priority=255
                    )
                    queue.declare(channel=channel)

    def get_plain_log_for_node(self, node_id: str, history_id: int = -1) -> str:
        """
        读取某个节点某一次执行的日志

        :param node_id: 节点 ID
        :type node_id: str
        :param history_id: 执行历史 ID, -1 表示获取最新日志
        :type history_id: int, optional
        :return: 节点日志
        :rtype: str
        """
        if history_id != -1:
            qs = ExecutionHistory.objects.filter(id=history_id).only("version")
        else:
            qs = State.objects.filter(node_id=node_id).only("version")

        if not qs:
            return ""
        version = qs.first().version
        return "\n".join(
            [
                e.message
                for e in LogEntry.objects.order_by("id").filter(node_id=node_id, version=version).only("message")
            ]
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json

from bamboo_engine import metrics
from bamboo_engine.eri import (
    Node,
    NodeType,
    ServiceActivity,
    SubProcess,
    ExclusiveGateway,
    ParallelGateway,
    ConditionalParallelGateway,
    ConvergeGateway,
    EmptyStartEvent,
    EmptyEndEvent,
    ExecutableEndEvent,
    Condition,
)

from pipeline.eri.models import Node as DBNode


class NodeMixin:
    def _get_node(self, node: DBNode):
        node_detail = json.loads(node.detail)
        node_type = node_detail["type"]
        targets = node_detail["targets"]
        common_args = dict(
            id=node.node_id,
            target_flows=list(targets.keys()),
            target_nodes=list(targets.values()),
            targets=node_detail["targets"],
            root_pipeline_id=node_detail["root_pipeline_id"],
            parent_pipeline_id=node_detail["parent_pipeline_id"],
            can_skip=node_detail["can_skip"],
            can_retry=node_detail["can_retry"],
        )

        if node_type == NodeType.ServiceActivity.value:
            return ServiceActivity(
                type=NodeType.ServiceActivity,
                code=node_detail["code"],
                version=node_detail["version"],
                timeout=node_detail["timeout"],
                error_ignorable=node_detail["error_ignorable"],
                **common_args
            )

        elif node_type == NodeType.SubProcess.value:
            return SubProcess(type=NodeType.SubProcess, start_event_id=node_detail["start_event_id"], **common_args)

        elif node_type == NodeType.ExclusiveGateway.value:
            return ExclusiveGateway(
                type=NodeType.ExclusiveGateway,
                conditions=[Condition(**c) for c in node_detail["conditions"]],
                **common_args
            )

        elif node_type == NodeType.ParallelGateway.value:
            return ParallelGateway(
                type=NodeType.ParallelGateway, converge_gateway_id=node_detail["converge_gateway_id"], **common_args
            )

        elif node_type == NodeType.ConditionalParallelGateway.value:
            return ConditionalParallelGateway(
                type=NodeType.ConditionalParallelGateway,
                converge_gateway_id=node_detail["converge_gateway_id"],
                conditions=[Condition(**c) for c in node_detail["conditions"]],
                **common_args
            )

        elif node_type == NodeType.ConvergeGateway.value:
            return ConvergeGateway(type=NodeType.ConvergeGateway, **common_args)

        elif node_type == NodeType.EmptyStartEvent.value:
            return EmptyStartEvent(type=NodeType.EmptyStartEvent, **common_args)

        elif node_type == NodeType.EmptyEndEvent.value:
            return EmptyEndEvent(type=NodeType.EmptyEndEvent, **common_args)

        elif node_type == NodeType.ExecutableEndEvent.value:
            return ExecutableEndEvent(type=NodeType.ExecutableEndEvent, code=node_detail["code"], **common_args)

        else:
            raise ValueError("unknown node type: {}".format(node_type))

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_NODE_READ_TIME)
    def get_node(self, node_id: str) -> Node:
        """
        获取某个节点的详细信息

        :param node_id: 节点 ID
        :type node_id: str
        :return: Node 实例
        :rtype: Node
        """
        node = DBNode.objects.get(node_id=node_id)
        return self._get_node(node)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import sys

from django.test import TestCase

from pipeline.contrib.external_plugins.tests.mock import *  # noqa
from pipeline.contrib.external_plugins.tests.mock_settings import *  # noqa
from pipeline.contrib.external_plugins.utils.importer import GitRepoModuleImporter, utils


class UtilsTestCase(TestCase):
    @patch(SYS_META_PATH, [])
    def test__set_up_importer(self):
        utils._setup_importer("1")
        utils._setup_importer("2")

        self.assertEqual(sys.meta_path, ["2", "1"])

    def test__remove_importer(self):
        importer_1 = GitRepoModuleImporter(
            name="name", modules=["module_1"], repo_raw_url="https://url_1", branch="master"
        )
        importer_2 = GitRepoModuleImporter(
            name="name", modules=["module_2"], repo_raw_url="https://url_2", branch="master"
        )
        importer_3 = GitRepoModuleImporter(
            name="name", modules=["module_3"], repo_raw_url="https://url_3", branch="master"
        )
        importer_4 = GitRepoModuleImporter(
            name="name", modules=["module_4"], repo_raw_url="https://url_4", branch="master"
        )

        with patch(SYS_META_PATH, [importer_1, importer_2, importer_3]):
            utils._remove_importer(importer_4)
            self.assertEqual(sys.meta_path, [importer_1, importer_2, importer_3])
            utils._remove_importer(importer_1)
            self.assertEqual(sys.meta_path, [importer_2, importer_3])
            utils._remove_importer(importer_3)
            self.assertEqual(sys.meta_path, [importer_2])
            utils._remove_importer(importer_2)
            self.assertEqual(sys.meta_path, [])

    @patch(UTILS_IMPORTER__SETUP_IMPORTER, MagicMock())
    @patch(UTILS_IMPORTER__REMOVE_IMPORTER, MagicMock())
    def test_importer_context__normal(self):
        importer = "importer"
        with utils.importer_context(importer):
            pass
        utils._setup_importer.assert_called_once_with(importer)
        utils._remove_importer.assert_called_once_with(importer)

    @patch(UTILS_IMPORTER__SETUP_IMPORTER, MagicMock())
    @patch(UTILS_IMPORTER__REMOVE_IMPORTER, MagicMock())
    def test_importer_context__raise_exception(self):
        importer = "importer"

        class CustomException(Exception):
            pass

        try:
            with utils.importer_context(importer):
                raise CustomException()
        except CustomException:
            pass

        utils._setup_importer.assert_called_once_with(importer)
        utils._remove_importer.assert_called_once_with(importer)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-10-28 02:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0025_auto_20200813_1216"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pipelinetemplate",
            name="name",
            field=models.CharField(db_index=True, default="default_template", max_length=128, verbose_name="模板名称"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from __future__ import absolute_import, unicode_literals

import os
import sys

import celery

try:
    import django_celery_beat
except ImportError:
    import djcelery

from kombu.utils.encoding import str_to_bytes
from django.core.management.base import BaseCommand

DB_SHARED_THREAD = """\
DatabaseWrapper objects created in a thread can only \
be used in that same thread.  The object with alias '{0}' \
was created in thread id {1} and this is thread id {2}.\
"""


def setenv(k, v):  # noqa
    os.environ[str_to_bytes(k)] = str_to_bytes(v)


def patch_thread_ident():
    # monkey patch django.
    # This patch make sure that we use real threads to get the ident which
    # is going to happen if we are using gevent or eventlet.
    # -- patch taken from gunicorn
    if getattr(patch_thread_ident, "called", False):
        return
    try:
        from django.db.backends.base.base import BaseDatabaseWrapper, DatabaseError

        if "validate_thread_sharing" in BaseDatabaseWrapper.__dict__:
            import threading

            _get_ident = threading.get_ident

            __old__init__ = BaseDatabaseWrapper.__init__

            def _init(self, *args, **kwargs):
                __old__init__(self, *args, **kwargs)
                self._thread_ident = _get_ident()

            def _validate_thread_sharing(self):
                if not self.allow_thread_sharing and self._thread_ident != _get_ident():
                    raise DatabaseError(DB_SHARED_THREAD % (self.alias, self._thread_ident, _get_ident()),)

            BaseDatabaseWrapper.__init__ = _init
            BaseDatabaseWrapper.validate_thread_sharing = _validate_thread_sharing

        patch_thread_ident.called = True
    except ImportError:
        pass


patch_thread_ident()


class CeleryCommand(BaseCommand):
    options = ()
    if hasattr(BaseCommand, "option_list"):
        options = BaseCommand.option_list
    else:

        def add_arguments(self, parser):
            option_typemap = {"string": str, "int": int, "float": float}
            for opt in self.option_list:
                option = {k: v for k, v in opt.__dict__.items() if v is not None}
                flags = option.get("_long_opts", []) + option.get("_short_opts", [])
                if option.get("default") == ("NO", "DEFAULT"):
                    option["default"] = None
                if option.get("nargs") == 1:
                    del option["nargs"]
                del option["_long_opts"]
                del option["_short_opts"]
                if "type" in option:
                    opttype = option["type"]
                    option["type"] = option_typemap.get(opttype, opttype)
                parser.add_argument(*flags, **option)

    skip_opts = ["--app", "--loader", "--config", "--no-color"]
    requires_system_checks = False
    keep_base_opts = False
    stdout, stderr = sys.stdout, sys.stderr

    def get_version(self):
        def get_version(self):
            try:
                version = "celery {c.__version__}\ndjango-celery-beat {d.__version__}".format(
                    c=celery, d=django_celery_beat,
                )
            except ImportError:
                version = "celery {c.__version__}\ndjango-celery {d.__version__}".format(c=celery, d=djcelery,)
            return version

    def execute(self, *args, **options):
        broker = options.get("broker")
        if broker:
            self.set_broker(broker)
        super(CeleryCommand, self).execute(*args, **options)

    def set_broker(self, broker):
        setenv("CELERY_BROKER_URL", broker)

    def run_from_argv(self, argv):
        self.handle_default_options(argv[2:])
        return super(CeleryCommand, self).run_from_argv(argv)

    def handle_default_options(self, argv):
        acc = []
        broker = None
        for i, arg in enumerate(argv):
            # --settings and --pythonpath are also handled
            # by BaseCommand.handle_default_options, but that is
            # called with the resulting options parsed by optparse.
            if "--settings=" in arg:
                _, settings_module = arg.split("=")
                setenv("DJANGO_SETTINGS_MODULE", settings_module)
            elif "--pythonpath=" in arg:
                _, pythonpath = arg.split("=")
                sys.path.insert(0, pythonpath)
            elif "--broker=" in arg:
                _, broker = arg.split("=")
            elif arg == "-b":
                broker = argv[i + 1]
            else:
                acc.append(arg)
        if broker:
            self.set_broker(broker)
        return argv if self.keep_base_opts else acc

    def die(self, msg):
        sys.stderr.write(msg)
        sys.stderr.write("\n")
        sys.exit()

    def _is_unwanted_option(self, option):
        return option._long_opts and option._long_opts[0] in self.skip_opts

    @property
    def option_list(self):
        return [x for x in self.options if not self._is_unwanted_option(x)]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from .conditional_parallel import ConditionalParallelGatewayHandler
from .converge_gateway import ConvergeGatewayHandler
from .empty_start_event import EmptyStartEventHandler
from .endevent import EmptyEndEventHandler, ExecutableEndEventHandler
from .exclusive_gateway import ExclusiveGatewayHandler
from .parallel_gateway import ParallelGatewayHandler
from .service_activity import ServiceActivityHandler
from .subprocess import SubprocessHandler


class HandlersFactory(object):
    _handlers = {
        EmptyStartEventHandler.element_cls(): EmptyStartEventHandler(),
        EmptyEndEventHandler.element_cls(): EmptyEndEventHandler(),
        ServiceActivityHandler.element_cls(): ServiceActivityHandler(),
        SubprocessHandler.element_cls(): SubprocessHandler(),
        ExclusiveGatewayHandler.element_cls(): ExclusiveGatewayHandler(),
        ParallelGatewayHandler.element_cls(): ParallelGatewayHandler(),
        ConditionalParallelGatewayHandler.element_cls(): ConditionalParallelGatewayHandler(),
        ConvergeGatewayHandler.element_cls(): ConvergeGatewayHandler(),
        ExecutableEndEventHandler.element_cls(): ExecutableEndEventHandler(),
    }

    _cluster_roots = [ExecutableEndEventHandler.element_cls()]

    @classmethod
    def find_cluster_root_cls(cls, element):
        for root in cls._cluster_roots:
            if issubclass(type(element), root):
                return root

        return type(element)

    @classmethod
    def handlers_for(cls, element):
        handler = cls._handlers.get(cls.find_cluster_root_cls(element))
        if not handler:
            raise KeyError("handler for element({element}) not found.".format(element=element))

        return handler



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pipeline.component_framework.base import ComponentMeta
from pipeline.core.data.base import DataObject
from pipeline.core.data.converter import get_variable
from pipeline.exceptions import ComponentDataLackException


class Component(object, metaclass=ComponentMeta):
    def __init__(self, data_dict):
        self.data_dict = data_dict

    @classmethod
    def outputs_format(cls):
        outputs = cls.bound_service().outputs()
        outputs = [oi.as_dict() for oi in outputs]
        return outputs

    @classmethod
    def inputs_format(cls):
        inputs = cls.bound_service().inputs()
        inputs = [ii.as_dict() for ii in inputs]
        return inputs

    @classmethod
    def _get_item_schema(cls, type, key):
        items = getattr(cls.bound_service(), type)()
        for item in items:
            if item.key == key:
                return item

        return None

    @classmethod
    def get_output_schema(cls, key):
        return cls._get_item_schema(type="outputs", key=key).schema

    @classmethod
    def get_input_schema(cls, key):
        return cls._get_item_schema(type="inputs", key=key).schema

    @classmethod
    def form_is_embedded(cls):
        return getattr(cls, "embedded_form", False)

    def clean_execute_data(self, context):
        """
        @summary: hook for subclass of Component to clean execute data with context
        @param context:
        @return:
        """
        return self.data_dict

    def data_for_execution(self, context, pipeline_data):
        data_dict = self.clean_execute_data(context)
        inputs = {}

        for key, tag_info in list(data_dict.items()):
            if tag_info is None:
                raise ComponentDataLackException("Lack of inputs: %s" % key)

            inputs[key] = get_variable(key, tag_info, context, pipeline_data)

        return DataObject(inputs)

    def service(self):
        return self.bound_service()



# Generated by Django 2.2.19 on 2021-03-29 12:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eri", "0002_auto_20210322_0233"),
    ]

    operations = [
        migrations.AddField(
            model_name="logentry",
            name="version",
            field=models.CharField(default="", max_length=33, verbose_name="状态版本"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback

from django.db import models
from django.utils.translation import ugettext_lazy as _

from pipeline.engine.conf import function_switch

logger = logging.getLogger("celery")


class FunctionSwitchManager(models.Manager):
    def init_db(self):
        try:
            name_set = {s.name for s in self.all()}
            s_to_be_created = []
            for switch in function_switch.switch_list:
                if switch["name"] not in name_set:
                    s_to_be_created.append(
                        FunctionSwitch(
                            name=switch["name"], description=switch["description"], is_active=switch["is_active"]
                        )
                    )
                else:
                    self.filter(name=switch["name"]).update(description=switch["description"])
            self.bulk_create(s_to_be_created)
        except Exception:
            logger.error("function switch init failed: %s" % traceback.format_exc())

    def is_frozen(self):
        return self.get(name=function_switch.FREEZE_ENGINE).is_active

    def freeze_engine(self):
        self.filter(name=function_switch.FREEZE_ENGINE).update(is_active=True)

    def unfreeze_engine(self):
        self.filter(name=function_switch.FREEZE_ENGINE).update(is_active=False)


class FunctionSwitch(models.Model):
    name = models.CharField(_("功能名称"), max_length=32, null=False, unique=True)
    description = models.TextField(_("功能描述"), default="")
    is_active = models.BooleanField(_("是否激活"), default=False)

    objects = FunctionSwitchManager()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from pipeline.conf import default_settings
from pipeline.core.data.hydration import hydrate_data, hydrate_node_data
from pipeline.core.flow.activity import SubProcess

from .base import FlowElementHandler

logger = logging.getLogger("pipeline_engine")

__all__ = ["SubprocessHandler"]


class SubprocessHandler(FlowElementHandler):
    @staticmethod
    def element_cls():
        return SubProcess

    def handle(self, process, element, status):
        # rerun mode
        if status.loop > 1:
            element.prepare_rerun_data()
            element.pipeline.context.recover_variable()
            process.top_pipeline.context.recover_variable()

        # set loop count
        element.data.outputs._loop = status.loop + default_settings.PIPELINE_RERUN_INDEX_OFFSET

        # pre output extract
        process.top_pipeline.context.extract_output(element, set_miss=False)

        # hydrate data
        hydrate_node_data(element)

        # context injection
        data = element.pipeline.data
        context = element.pipeline.context
        for k, v in list(data.get_inputs().items()):
            context.set_global_var(k, v)

        hydrated = hydrate_data(context.variables)
        context.update_global_var(hydrated)

        sub_pipeline = element.pipeline
        process.push_pipeline(sub_pipeline, is_subprocess=True)
        process.take_snapshot()
        return self.HandleResult(next_node=sub_pipeline.start_event, should_return=False, should_sleep=False)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-13 07:43
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("component_framework", "0005_auto_20190723_1806"),
    ]

    operations = [
        migrations.AlterField(
            model_name="componentmodel",
            name="version",
            field=models.CharField(default="legacy", max_length=64, verbose_name="组件版本"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import models
from django.utils.translation import ugettext_lazy as _


class Process(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    parent_id = models.BigIntegerField(_("父进程 ID"), default=-1, db_index=True)
    ack_num = models.IntegerField(_("收到子进程 ACK 数量"), default=0)
    need_ack = models.IntegerField(_("需要收到的子进程 ACK 数量"), default=-1)
    asleep = models.BooleanField(_("是否处于休眠状态"), default=True)
    suspended = models.BooleanField(_("是否处于暂停状态"), default=False)
    frozen = models.BooleanField(_("是否处于冻结状态"), default=False)
    dead = models.BooleanField(_("是否已经死亡"), default=False)
    last_heartbeat = models.DateTimeField(_("上次心跳时间"), auto_now_add=True, db_index=True)
    destination_id = models.CharField(_("执行终点 ID"), default="", max_length=33)
    current_node_id = models.CharField(_("当前节点 ID"), default="", max_length=33, db_index=True)
    root_pipeline_id = models.CharField(_("根流程 ID"), null=False, max_length=33, db_index=True)
    suspended_by = models.CharField(_("导致进程暂停的节点 ID"), default="", max_length=33, db_index=True)
    priority = models.IntegerField(_("优先级"))
    queue = models.CharField(_("所属队列"), default="", max_length=128)
    pipeline_stack = models.TextField(_("流程栈"), default="[]", null=False)


class Node(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), null=False, max_length=33, db_index=True)
    detail = models.TextField(_("节点详情"), null=False)


class State(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), null=False, max_length=33, unique=True)
    root_id = models.CharField(_("根节点 ID"), null=False, default="", max_length=33, db_index=True)
    parent_id = models.CharField(_("父节点 ID"), null=False, default="", max_length=33, db_index=True)
    name = models.CharField(_("状态名"), null=False, max_length=64)
    version = models.CharField(_("状态版本"), null=False, max_length=33)
    loop = models.IntegerField(_("循环次数"), default=1)
    retry = models.IntegerField(_("重试次数"), default=0)
    skip = models.BooleanField(_("是否跳过"), default=False)
    error_ignored = models.BooleanField(_("是否出错后自动忽略"), default=False)
    created_time = models.DateTimeField(_("创建时间"), auto_now_add=True)
    started_time = models.DateTimeField(_("开始时间"), null=True)
    archived_time = models.DateTimeField(_("归档时间"), null=True)


class Schedule(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    type = models.IntegerField(_("调度类型"))
    process_id = models.BigIntegerField(_("进程 ID"), default=-1)
    node_id = models.CharField(_("节点 ID"), null=False, max_length=33)
    finished = models.BooleanField(_("是否已完成"), default=False)
    expired = models.BooleanField(_("是否已过期"), default=False)
    scheduling = models.BooleanField(_("是否正在调度"), default=False)
    version = models.CharField(_("状态版本"), null=False, max_length=33)
    schedule_times = models.IntegerField(_("被调度次数"), default=0)

    class Meta:
        unique_together = ["node_id", "version"]


class Data(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), null=False, max_length=33, db_index=True, unique=True)
    inputs = models.TextField(_("原始输入数据"))
    outputs = models.TextField(_("原始输出数据"))


class ExecutionData(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), null=False, max_length=33, db_index=True, unique=True)
    inputs_serializer = models.CharField(_("输入序列化器"), null=False, max_length=32)
    outputs_serializer = models.CharField(_("输出序列化器"), null=False, max_length=32)
    inputs = models.TextField(_("节点执行输入数据"))
    outputs = models.TextField(_("节点执行输出数据"))


class CallbackData(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), null=False, max_length=33)
    version = models.CharField(_("状态版本"), null=False, max_length=33)
    data = models.TextField(_("回调数据"))


class ContextValue(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    pipeline_id = models.CharField(_("流程 ID"), null=False, max_length=33)
    key = models.CharField(_("变量 key"), null=False, max_length=128)
    type = models.IntegerField(_("变量类型"))
    serializer = models.CharField(_("序列化器"), null=False, max_length=32)
    code = models.CharField(_("计算型变量类型唯一标志"), default="", max_length=128)
    value = models.TextField(_("变量值"))
    references = models.TextField(_("所有对其他变量直接或间接的引用"))

    class Meta:
        unique_together = ["pipeline_id", "key"]


class ContextOutputs(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    pipeline_id = models.CharField(_("流程 ID"), null=False, max_length=33, unique=True)
    outputs = models.TextField(_("输出配置"))


class ExecutionHistory(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), null=False, max_length=33)
    loop = models.IntegerField(_("循环次数"), default=1)
    retry = models.IntegerField(_("重试次数"), default=0)
    skip = models.BooleanField(_("是否跳过"), default=False)
    version = models.CharField(_("状态版本"), null=False, max_length=33)
    started_time = models.DateTimeField(_("开始时间"), null=False)
    archived_time = models.DateTimeField(_("归档时间"), null=False)
    inputs_serializer = models.CharField(_("输入序列化器"), null=False, max_length=32)
    outputs_serializer = models.CharField(_("输出序列化器"), null=False, max_length=32)
    inputs = models.TextField(_("节点执行输入数据"))
    outputs = models.TextField(_("节点执行输出数据"))

    class Meta:
        index_together = ["node_id", "loop"]


class LogEntry(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), max_length=33)
    version = models.CharField(_("状态版本"), default="", max_length=33)
    loop = models.IntegerField(_("循环次数"), default=1)
    logger_name = models.CharField(_("logger 名称"), max_length=128)
    level_name = models.CharField(_("日志等级"), max_length=32)
    message = models.TextField(_("日志内容"), null=True)
    logged_at = models.DateTimeField(_("输出时间"), auto_now_add=True, db_index=True)

    class Meta:
        index_together = ["node_id", "loop"]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import sys
import logging
import traceback

import redis
from django.apps import AppConfig
from django.conf import settings
from redis.sentinel import Sentinel
from rediscluster import RedisCluster

logger = logging.getLogger("root")


def get_client_through_sentinel():
    kwargs = {"sentinel_kwargs": {}}
    sentinel_pwd = settings.REDIS.get("sentinel_password")
    if sentinel_pwd:
        kwargs["sentinel_kwargs"]["password"] = sentinel_pwd
    if "password" in settings.REDIS:
        kwargs["password"] = settings.REDIS["password"]
    host = settings.REDIS["host"]
    port = settings.REDIS["port"]
    sentinels = list(zip([h.strip() for h in host.split(",")], [p.strip() for p in str(port).split(",")],))
    rs = Sentinel(sentinels, **kwargs)
    # avoid None value in settings.REDIS
    r = rs.master_for(settings.REDIS.get("service_name") or "mymaster")
    # try to connect master
    r.echo("Hello Redis")
    return r


def get_cluster_client():
    kwargs = {"startup_nodes": [{"host": settings.REDIS["host"], "port": settings.REDIS["port"]}]}
    if "password" in settings.REDIS:
        kwargs["password"] = settings.REDIS["password"]

    r = RedisCluster(**kwargs)
    r.echo("Hello Redis")
    return r


def get_single_client():
    kwargs = {
        "host": settings.REDIS["host"],
        "port": settings.REDIS["port"],
    }
    if "password" in settings.REDIS:
        kwargs["password"] = settings.REDIS["password"]
    if "db" in settings.REDIS:
        kwargs["db"] = settings.REDIS["db"]

    pool = redis.ConnectionPool(**kwargs)
    return redis.StrictRedis(connection_pool=pool)


CLIENT_GETTER = {
    "replication": get_client_through_sentinel,
    "cluster": get_cluster_client,
    "single": get_single_client,
}


class PipelineConfig(AppConfig):
    name = "pipeline"
    verbose_name = "Pipeline"

    def ready(self):
        from pipeline.signals.handlers import pipeline_template_post_save_handler  # noqa
        from pipeline.validators.handlers import post_new_end_event_register_handler  # noqa

        # init redis pool
        if hasattr(settings, "REDIS"):
            mode = settings.REDIS.get("mode") or "single"
            try:
                settings.redis_inst = CLIENT_GETTER[mode]()
            except Exception:
                # fall back to single node mode
                logger.error("redis client init error: %s" % traceback.format_exc())
        elif (
            getattr(settings, "PIPELINE_DATA_BACKEND", None)
            == "pipeline.engine.core.data.redis_backend.RedisDataBackend"
        ):
            logger.error("can not find REDIS in settings!")

        # avoid big flow pickle raise maximum recursion depth exceeded error
        sys.setrecursionlimit(10000)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
from abc import abstractmethod

from pipeline import exceptions
from pipeline.core.data import library
from pipeline.core.data.context import OutputRef
from pipeline.core.data.expression import ConstantTemplate, format_constant_key
from pipeline.core.signals import pre_variable_register

logger = logging.getLogger("root")


class Variable(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    @abstractmethod
    def get(self):
        pass


class PlainVariable(Variable):
    def __init__(self, name, value):
        super(PlainVariable, self).__init__(name, value)
        self.name = name
        self.value = value

    def get(self):
        return self.value

    def __repr__(self):
        return "[plain_var] {}".format(self.name)

    def __str__(self):
        return self.__repr__()

    def __unicode__(self):
        return self.__repr__()


class SpliceVariable(Variable):
    def __init__(self, name, value, context):
        super(SpliceVariable, self).__init__(name, value)
        self._value = None
        self._build_reference(context)

    def get(self):
        if not self._value:
            try:
                self._resolve()
            except Exception as e:
                logger.error("get value[{}] of Variable[{}] error[{}]".format(self.value, self.name, e))
                return self.value
        return self._value

    def _build_reference(self, context):
        keys = ConstantTemplate(self.value).get_reference()
        refs = {}
        for key in keys:
            refs[key] = OutputRef(format_constant_key(key), context)
        self._refs = refs

    def _resolve(self):
        maps = {}
        for key in self._refs:
            try:
                ref_val = self._refs[key].value
                if issubclass(ref_val.__class__, Variable):
                    ref_val = ref_val.get()
            except exceptions.ReferenceNotExistError:
                continue
            maps[key] = ref_val
        val = ConstantTemplate(self.value).resolve_data(maps)

        self._value = val

    def __repr__(self):
        return "[splice_var] {}".format(self.name)

    def __str__(self):
        return self.__repr__()

    def __unicode__(self):
        return self.__repr__()


class RegisterVariableMeta(type):
    def __new__(cls, name, bases, attrs):
        super_new = super(RegisterVariableMeta, cls).__new__

        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        parents = [b for b in bases if isinstance(b, RegisterVariableMeta)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Create the class
        new_class = super_new(cls, name, bases, attrs)

        if not new_class.code:
            raise exceptions.ConstantReferenceException("LazyVariable %s: code can't be empty." % new_class.__name__)

        pre_variable_register.send(sender=LazyVariable, variable_cls=new_class)

        library.VariableLibrary.variables[new_class.code] = new_class

        return new_class


class LazyVariable(SpliceVariable, metaclass=RegisterVariableMeta):
    def __init__(self, name, value, context, pipeline_data):
        super(LazyVariable, self).__init__(name, value, context)
        self.context = context
        self.pipeline_data = pipeline_data

    # variable reference resolve
    def get(self):
        self.value = super(LazyVariable, self).get()
        try:
            return self.get_value()
        except Exception as e:
            logger.error("get value[{}] of Variable[{}] error[{}]".format(self.value, self.name, e))
            return self.value

    # get real value by user code
    @abstractmethod
    def get_value(self):
        pass



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import ujson as json
import contextlib
import logging
import traceback

from celery import current_app
from celery.task.control import revoke
from django.db import models, transaction
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from pipeline.conf import settings as pipeline_settings
from pipeline.constants import PIPELINE_DEFAULT_PRIORITY
from pipeline.core.data.base import DataObject
from pipeline.core.pipeline import Pipeline
from pipeline.django_signal_valve import valve
from pipeline.engine import exceptions, signals, states, utils
from pipeline.engine.core import data as data_service
from pipeline.engine.models.fields import IOField
from pipeline.engine.utils import ActionResult, Stack, calculate_elapsed_time
from pipeline.log.models import LogEntry
from pipeline.utils.uniqid import node_uniqid, uniqid

logger = logging.getLogger("celery")

RERUN_MAX_LIMIT = pipeline_settings.PIPELINE_RERUN_MAX_TIMES
NAME_MAX_LENGTH = 64


class ProcessSnapshotManager(models.Manager):
    def create_snapshot(self, pipeline_stack, children, root_pipeline, subprocess_stack):
        data = {
            "_pipeline_stack": pipeline_stack,
            "_subprocess_stack": subprocess_stack,
            "_children": children,
            "_root_pipeline": root_pipeline,
        }
        return self.create(data=data)


class ProcessSnapshot(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    data = IOField(verbose_name=_("pipeline 运行时数据"))

    objects = ProcessSnapshotManager()

    @property
    def pipeline_stack(self):
        return self.data["_pipeline_stack"]

    @property
    def children(self):
        return self.data["_children"]

    @property
    def root_pipeline(self):
        return self.data["_root_pipeline"]

    @property
    def subprocess_stack(self):
        return self.data["_subprocess_stack"]

    def clean_children(self):
        self.data["_children"] = []

    def prune_top_pipeline(self, keep_from, keep_to):
        self.data["_pipeline_stack"].top().prune(keep_from, keep_to)


class ProcessManager(models.Manager):
    def prepare_for_pipeline(self, pipeline):
        """
        为 pipeline 创建相应的 process 并进行一系列初始化
        :param pipeline:
        :return:
        """
        # init runtime info
        snapshot = ProcessSnapshot.objects.create_snapshot(
            pipeline_stack=utils.Stack(), children=[], root_pipeline=pipeline, subprocess_stack=utils.Stack(),
        )
        process = self.create(
            id=node_uniqid(), root_pipeline_id=pipeline.id, current_node_id=pipeline.start_event.id, snapshot=snapshot,
        )
        process.push_pipeline(pipeline)
        process.save()
        return process

    def fork_child(self, parent, current_node_id, destination_id):
        """
        创建一个上下文信息与当前 parent 一致的 child process
        :param parent:
        :param current_node_id:
        :param destination_id:
        :return:
        """
        # init runtime info

        # clear parent's change
        parent.top_pipeline.context.clear_change_keys()

        pipeline_stack = Stack([parent.top_pipeline])
        root_pipeline_shell = parent.root_pipeline.shell()
        snapshot = ProcessSnapshot.objects.create_snapshot(
            pipeline_stack=pipeline_stack,
            children=[],
            root_pipeline=root_pipeline_shell,
            subprocess_stack=parent.subprocess_stack,
        )
        # refresh first, avoid keep the same ref to parent.top_pipeline
        snapshot.refresh_from_db()
        snapshot.prune_top_pipeline(current_node_id, destination_id)
        snapshot.save()

        child = self.create(
            id=node_uniqid(),
            root_pipeline_id=parent.root_pipeline.id,
            current_node_id=current_node_id,
            destination_id=destination_id,
            parent_id=parent.id,
            snapshot=snapshot,
        )
        for subproc_id in parent.subprocess_stack:
            SubProcessRelationship.objects.add_relation(subproc_id, child.id)

        return child

    def process_ready(self, process_id, current_node_id=None, call_from_child=False):
        """
        发送一个进程已经准备好被调度的信号
        :param process_id: 已经准备好的进程 ID
        :param current_node_id: 下一个执行的节点的 ID（可用于失败跳过）
        :param call_from_child: 该信号是否由子进程发出
        :return:
        """
        valve.send(
            signals,
            "process_ready",
            sender=PipelineProcess,
            process_id=process_id,
            current_node_id=current_node_id,
            call_from_child=call_from_child,
        )

    def batch_process_ready(self, process_id_list, pipeline_id):
        """
        发送批量唤醒进程的信号
        :param process_id_list: 需要被唤醒的进程 ID 列表
        :param pipeline_id: 这些进程相关的 root pipeline
        :return:
        """
        valve.send(
            signals,
            "batch_process_ready",
            sender=PipelineProcess,
            process_id_list=process_id_list,
            pipeline_id=pipeline_id,
        )

    def child_process_ready(self, child_id):
        """
        发送子进程已经准备好被调度的信号
        :param child_id: 子进程 ID
        :return:
        """
        valve.send(signals, "child_process_ready", sender=PipelineProcess, child_id=child_id)

    def priority_for_process(self, process_id):
        """
        查询进程对应的 pipeline 的优先级
        :param process_id: 进程 ID
        :return:
        """
        return PipelineModel.objects.get(id=self.get(id=process_id).root_pipeline_id).priority

    def queue_for_process(self, process_id):
        """
        查询进程对应的 pipeline 所使用的队列
        :param process_id: 进程 ID
        :return:
        """
        return PipelineModel.objects.get(id=self.get(id=process_id).root_pipeline_id).queue

    def task_args_for_process(self, process_id):
        pipeline_model = PipelineModel.objects.get(id=self.get(id=process_id).root_pipeline_id)

        return {"priority": pipeline_model.priority, "queue": pipeline_model.queue}


class PipelineProcess(models.Model):
    """
    @relationship with other models
        1. PipelineInstance
            process = PipelineProcess.objects.get(root_pipeline_id=pipeline_inst.instance_id)
            pipeline_inst = PipelineInstance.objects.get(instance_id=process.root_pipeline_id)
    """

    id = models.CharField(_("Process ID"), unique=True, primary_key=True, max_length=32)
    root_pipeline_id = models.CharField(_("根 pipeline 的 ID"), max_length=32, db_index=True)
    current_node_id = models.CharField(_("当前推进到的节点的 ID"), max_length=32, default="", db_index=True)
    destination_id = models.CharField(_("遇到该 ID 的节点就停止推进"), max_length=32, default="")
    parent_id = models.CharField(_("父 process 的 ID"), max_length=32, default="")
    ack_num = models.IntegerField(_("收到子节点 ACK 的数量"), default=0)
    need_ack = models.IntegerField(_("需要收到的子节点 ACK 的数量"), default=-1)
    is_alive = models.BooleanField(_("该 process 是否还有效"), default=True, db_index=True)
    is_sleep = models.BooleanField(_("该 process 是否正在休眠"), default=False, db_index=True)
    is_frozen = models.BooleanField(_("该 process 是否被冻结"), default=False, db_index=True)
    snapshot = models.ForeignKey(ProcessSnapshot, null=True, on_delete=models.SET_NULL)

    objects = ProcessManager()

    @property
    def pipeline_stack(self):
        return self.snapshot.pipeline_stack if self.snapshot else None

    @property
    def children(self):
        return self.snapshot.children if self.snapshot else None

    @property
    def root_pipeline(self):
        return self.snapshot.root_pipeline if self.snapshot else None

    @property
    def top_pipeline(self):
        return self.pipeline_stack.top()

    @property
    def subprocess_stack(self):
        return self.snapshot.subprocess_stack if self.snapshot else None

    @property
    def in_subprocess(self):
        return len(self.snapshot.pipeline_stack) > 1 if self.snapshot else False

    def push_pipeline(self, pipeline, is_subprocess=False):
        """
        将 pipeline 压入运行时栈中
        :param pipeline: 需要被压入栈中的 pipeline 对象
        :param is_subprocess: 该 pipeline 是否是子流程
        :return:
        """
        self.pipeline_stack.push(pipeline)
        if is_subprocess:
            self.subprocess_stack.push(pipeline.id)
            SubProcessRelationship.objects.add_relation(pipeline.id, self.id)

    def pop_pipeline(self):
        """
        从运行时栈中弹出一个 pipeline
        :return:
        """
        pipeline = self.pipeline_stack.pop()
        if self.subprocess_stack:
            subproc_id = self.subprocess_stack.pop()
            SubProcessRelationship.objects.delete_relation(subproc_id, self.id)
        return pipeline

    def join(self, children):
        """
        令父进程等待子进程
        :param children: 需要等待的子进程列表
        :return:
        """
        self.need_ack = len(children)
        for child in children:
            self.children.append(child.id)
        self.save()

    def root_sleep_check(self):
        """
        检测 root pipeline 的状态判断当前进程是否需要休眠
        :return:
        """
        root_state = Status.objects.state_for(self.root_pipeline.id)
        if root_state in states.SLEEP_STATES:
            return True, root_state
        if root_state == states.BLOCKED:
            # 子进程不会因为父进程进入了 BLOCKED 状态就进入睡眠
            return not self.parent_id, root_state
        return False, root_state

    def subproc_sleep_check(self):
        """
        检测当前子流程栈中所有子流程的状态判断当前进程是否需要休眠
        :return:
        """
        status = Status.objects.filter(id__in=self.subprocess_stack)
        status_map = {s.id: s.state for s in status}
        # 记录第一个处于暂停状态之前的所有子流程，用于子流程状态的修改
        before_suspended = []
        for subproc_id in self.subprocess_stack:
            if status_map[subproc_id] == states.SUSPENDED:
                return True, before_suspended
            else:
                before_suspended.append(subproc_id)
        return False, before_suspended

    def freeze(self):
        """
        冻结当前进程
        :return:
        """
        with transaction.atomic():
            self.__class__.objects.select_for_update().get(id=self.id)
            self.is_frozen = True
            self.save()
            ProcessCeleryTask.objects.unbind(self.id)

    def unfreeze(self):
        """
        解冻当前进程
        :return:
        """
        with transaction.atomic():
            self.__class__.objects.select_for_update().get(id=self.id)
            self.is_frozen = False
            self.save(save_snapshot=False)
            valve.send(signals, "process_unfreeze", sender=PipelineProcess, process_id=self.id)

    def sleep(self, do_not_save=False, adjust_status=False, adjust_scope=None):
        """
        休眠当前进程
        :param do_not_save: 是否需要保存进程信息
        :param adjust_status: 是否需要调整 pipeline 中当前节点父级节点的状态
        :param adjust_scope: 状态调整的范围
        :return:
        """
        if adjust_status:
            self.adjust_status(adjust_scope)
        if do_not_save:
            return
        with transaction.atomic():
            self.__class__.objects.select_for_update().get(id=self.id)
            self.is_sleep = True
            self.save()
            ProcessCeleryTask.objects.unbind(self.id)
        # dispatch children
        for child_id in self.children:
            PipelineProcess.objects.child_process_ready(child_id)

    def adjust_status(self, adjust_scope=None):
        """
        根据当前节点和子流程的状态来调整父级节点的状态
        :param adjust_scope: 子流程状态调整范围
        :return:
        """
        node_state = Status.objects.state_for(self.current_node_id, may_not_exist=True)
        pipeline_state = Status.objects.state_for(self.root_pipeline.id, may_not_exist=True)
        subproc_states = Status.objects.states_for(self.subprocess_stack)

        if node_state in {states.FAILED, states.SUSPENDED}:
            # if current node failed or suspended
            Status.objects.batch_transit(
                id_list=self.subprocess_stack, state=states.BLOCKED, from_state=states.RUNNING,
            )
            Status.objects.transit(self.root_pipeline.id, to_state=states.BLOCKED, is_pipeline=True)
        elif states.SUSPENDED in set(subproc_states):
            # if any subprocess suspended
            Status.objects.batch_transit(id_list=adjust_scope, state=states.BLOCKED, from_state=states.RUNNING)
            Status.objects.transit(self.root_pipeline.id, to_state=states.BLOCKED, is_pipeline=True)
        elif pipeline_state == states.SUSPENDED:
            # if root pipeline suspended
            Status.objects.batch_transit(
                id_list=self.subprocess_stack, state=pipeline_state, from_state=states.RUNNING,
            )

    def wake_up(self):
        """
        唤醒当前进程
        :return:
        """
        with transaction.atomic():
            self.__class__.objects.select_for_update().get(id=self.id)
            self.is_sleep = False
            self.save(save_snapshot=False)

    def destroy(self):
        """
        销毁当前进程及其上下文数据
        :return:
        """
        self.is_alive = False
        self.current_node_id = ""
        snapshot = self.snapshot
        self.snapshot = None

        self.save()
        snapshot.delete()
        ProcessCeleryTask.objects.destroy(self.id)

    def destroy_all(self):
        """
        销毁当前进程并递归销毁其所有子进程
        :return:
        """
        _destroy_recursively(self)

    def save(self, save_snapshot=True, **kwargs):
        if save_snapshot and self.snapshot:
            self.snapshot.save()
        return super(PipelineProcess, self).save(**kwargs)

    def blocked_by_failure_or_suspended(self):
        """
        检测当前进程是否因为节点失败而休眠
        :return:
        """
        if not self.is_sleep:
            return False
        if Status.objects.state_for(self.current_node_id, may_not_exist=True) in {
            states.FAILED,
            states.SUSPENDED,
        }:
            return True
        if not self.children:
            return False
        children = self.__class__.objects.filter(id__in=self.children)
        result = []
        for child in children:
            result.append(child.blocked_by_failure_or_suspended())
        return True in result

    def sync_with_children(self):
        """
        与子进程同步数据
        :return:
        """
        for child_id in self.children:
            context = data_service.get_object(self._context_key(child_id))
            parent_data = data_service.get_object(self._data_key(child_id))
            if context is None or parent_data is None:
                raise exceptions.ChildDataSyncError(
                    "sync data with children {} failed, context({}) or parent_data({}) is None".format(
                        child_id, context, parent_data
                    )
                )
            self.top_pipeline.context.sync_change(context)
            # self.top_pipeline.context.update_global_var(context.variables)
            self.top_pipeline.data.update_outputs(parent_data.get_outputs())
        self.clean_children()  # remove all children

    def destroy_and_wake_up_parent(self, destination_id):
        """
        销毁当前进程并尝试唤醒父进程
        :param destination_id: 当前进程终点节点 ID
        :return:
        """
        # save sync data
        data_service.set_object(self._context_key(), self.top_pipeline.context)
        data_service.set_object(self._data_key(), self.top_pipeline.data)

        self.__class__.objects.filter(id=self.parent_id).update(ack_num=models.F("ack_num") + 1)
        can_wake_up = False

        with transaction.atomic():
            parent = self.__class__.objects.select_for_update().get(id=self.parent_id)

            if parent.need_ack != -1:
                if parent.ack_num == parent.need_ack:
                    # try to wake up parent
                    parent.need_ack = -1
                    parent.ack_num = 0
                    can_wake_up = True
                else:
                    if parent.blocked_by_failure_or_suspended():
                        Status.objects.batch_transit(
                            id_list=self.subprocess_stack, state=states.BLOCKED, from_state=states.RUNNING,
                        )
                        Status.objects.transit(
                            id=self.root_pipeline.id, to_state=states.BLOCKED, is_pipeline=True,
                        )

            parent.save(save_snapshot=False)

        if can_wake_up:
            self.__class__.objects.process_ready(parent.id, current_node_id=destination_id, call_from_child=True)

        SubProcessRelationship.objects.delete_relation(None, self.id)
        self.destroy()

    def _context_key(self, process_id=None):
        return "%s_context" % (process_id if process_id else self.id)

    def _data_key(self, process_id=None):
        return "%s_data" % (process_id if process_id else self.id)

    def can_be_waked(self):
        """
        检测当前进程是否能够被唤醒
        :return:
        """
        if not self.is_sleep or not self.is_alive:
            return False
        if self.need_ack != -1 and self.need_ack != self.ack_num:
            return False
        return True

    def clean_children(self):
        """
        清空当前进程的 children
        :return:
        """
        for child_id in self.children:
            # 删除子进程的数据
            data_service.del_object(self._context_key(child_id))
            data_service.del_object(self._data_key(child_id))

        self.snapshot.clean_children()
        self.snapshot.save()

    def exit_gracefully(self, e):
        """
        在遇到无法处理的异常时优雅的退出当前进程
        :param e:
        :return:
        """
        ex_data = traceback.format_exc()
        try:
            current_node = self.top_pipeline.node(self.current_node_id)
        except IndexError:
            current_node = self.root_pipeline.node(self.current_node_id)

        # current_node may be a end_event in pipeline which had been popped
        if current_node is not None:
            result = Status.objects.fail(current_node, ex_data=ex_data)
        else:
            result = Status.objects.raw_fail(self.current_node_id, ex_data=ex_data)

        if not result.result:
            logger.error(
                "process({process_id}) exit_gracefully status transit failed, current_node :{node_id}".format(
                    process_id=self.id, node_id=current_node.id if current_node else self.current_node_id,
                )
            )
        self.sleep(adjust_status=True)

    def refresh_current_node(self, current_node_id):
        """
        刷新当前节点的 ID
        :param current_node_id:
        :return:
        """
        self.__class__.objects.filter(id=self.id).update(current_node_id=current_node_id)

    def revoke_subprocess(self):
        if self.subprocess_stack:
            Status.objects.batch_transit(id_list=list(self.subprocess_stack), state=states.REVOKED)

        if self.children:
            for child_id in self.children:
                PipelineProcess.objects.get(id=child_id).revoke_subprocess()

    def take_snapshot(self):
        """
        保存当前进程的快照对象
        :return:
        """
        self.snapshot.save()


def _destroy_recursively(process):
    if not process.is_alive:
        return
    if process.children:
        for child_id in process.children:
            child = PipelineProcess.objects.get(id=child_id)
            _destroy_recursively(child)
        process.destroy()
    else:
        process.destroy()


class PipelineModelManager(models.Manager):
    def prepare_for_pipeline(self, pipeline, process, priority, queue=""):
        return self.create(id=pipeline.id, process=process, priority=priority, queue=queue)

    def pipeline_ready(self, process_id):
        valve.send(signals, "pipeline_ready", sender=Pipeline, process_id=process_id)

    def priority_for_pipeline(self, pipeline_id):
        return self.get(id=pipeline_id).priority

    def task_args_for_pipeline(self, pipeline_id):
        model = self.get(id=pipeline_id)

        return {"priority": model.priority, "queue": model.queue}


class PipelineModel(models.Model):
    id = models.CharField("pipeline ID", unique=True, primary_key=True, max_length=32)
    process = models.ForeignKey(PipelineProcess, null=True, on_delete=models.SET_NULL)
    priority = models.IntegerField(_("流程优先级"), default=PIPELINE_DEFAULT_PRIORITY)
    queue = models.CharField(_("流程使用的队列名"), max_length=512, default="")

    objects = PipelineModelManager()


class RelationshipManager(models.Manager):
    def build_relationship(self, ancestor_id, descendant_id):
        if self.filter(ancestor_id=ancestor_id, descendant_id=descendant_id).exists():
            # already build
            return
        ancestors = self.filter(descendant_id=ancestor_id)
        relationships = [NodeRelationship(ancestor_id=descendant_id, descendant_id=descendant_id, distance=0)]
        for ancestor in ancestors:
            rel = NodeRelationship(
                ancestor_id=ancestor.ancestor_id, descendant_id=descendant_id, distance=ancestor.distance + 1,
            )
            relationships.append(rel)
        self.bulk_create(relationships)


class NodeRelationship(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    ancestor_id = models.CharField(_("祖先 ID"), max_length=32, db_index=True)
    descendant_id = models.CharField(_("后代 ID"), max_length=32, db_index=True)
    distance = models.IntegerField(_("距离"), db_index=True)

    objects = RelationshipManager()

    def __unicode__(self):
        return str("#{} -({})-> #{}".format(self.ancestor_id, self.distance, self.descendant_id,))


class StatusManager(models.Manager):
    def transit(
        self, id, to_state, is_pipeline=False, appoint=False, start=False, name="", version=None, unchanged_pass=False,
    ):
        """
        尝试改变某个节点的状态
        :param id: 节点 ID
        :param to_state: 目标状态
        :param is_pipeline: 该节点是否是 pipeline
        :param appoint: 该动作是否由用户发起（非引擎内部操作）
        :param start: 是否刷新其开始时间
        :param name: 节点名称
        :param version: 节点版本
        :param unchanged_pass: 当 to_state 与当前节点状态相同时则视为操作成功
        :return:
        """
        defaults = {
            "name": name,
            "state": to_state,
            "version": uniqid(),
        }
        if start:
            now = timezone.now()
            defaults["started_time"] = now
            defaults["state_refresh_at"] = now
        status, created = self.get_or_create(id=id, defaults=defaults)

        # reservation or first creation
        if created:
            return ActionResult(result=True, message="success", extra=status)

        with transaction.atomic():
            kwargs = {"id": id}
            if version:
                kwargs["version"] = version

            try:
                status = self.select_for_update().get(**kwargs)

            except Status.DoesNotExist:
                return ActionResult(result=False, message="node not exists or not be executed yet")

            if unchanged_pass and status.state == to_state:
                return ActionResult(result=True, message="success", extra=status)

            if states.can_transit(
                from_state=status.state, to_state=to_state, is_pipeline=is_pipeline, appoint=appoint,
            ):

                # 在冻结状态下不能改变 pipeline 的状态
                if is_pipeline:
                    subprocess_rel = SubProcessRelationship.objects.filter(subprocess_id=id)
                    if subprocess_rel:
                        process = PipelineProcess.objects.get(id=subprocess_rel[0].process_id)
                        if process.is_frozen:
                            return ActionResult(result=False, message="engine is frozen, can not perform operation",)

                    processes = PipelineProcess.objects.filter(root_pipeline_id=id)
                    if processes and processes[0].is_frozen:
                        return ActionResult(result=False, message="engine is frozen, can not perform operation",)

                if name:
                    status.name = name
                if to_state in states.ARCHIVED_STATES:
                    status.archived_time = timezone.now()

                # from FINISHED to RUNNING
                if states.is_rerunning(from_state=status.state, to_state=to_state):
                    history = History.objects.record(status, is_rerunning=True)
                    if history:
                        LogEntry.objects.link_history(node_id=status.id, history_id=history.id)
                    status.loop += 1
                    status.skip = False
                    status.version = uniqid()

                # reset started_time after record last status
                if start:
                    status.started_time = timezone.now()
                status.state = to_state
                status.state_refresh_at = timezone.now()
                status.save()
                return ActionResult(result=True, message="success", extra=status)
            else:
                return ActionResult(
                    result=False,
                    message="can't transit state({}) from {} to {}".format(id, status.state, to_state),
                    extra=status,
                )

    def batch_transit(self, id_list, state, from_state=None, exclude=None):
        """
        批量改变节点状态，仅用于子流程的状态修改
        :param id_list: 子流程 ID 列表
        :param state: 目标状态
        :param from_state: 起始状态
        :param exclude: 不需要改变状态的子流程 ID 列表
        :return:
        """
        if not id_list:
            return
        if not exclude:
            exclude = []

        id_list = set(id_list)
        exclude = set(exclude)
        kwargs = {"id__in": [i for i in id_list if i not in exclude]}
        if from_state:
            kwargs["state"] = from_state
        with transaction.atomic():
            self.select_for_update().filter(**kwargs).update(state=state)

    def state_for(self, id, may_not_exist=False, version=None):
        """
        获取某个节点的状态
        :param id: 节点 ID
        :param may_not_exist: 该节点是否可能不存在（未执行到）
        :param version: 节点版本
        :return:
        """
        kwargs = {"id": id}
        if version:
            kwargs["version"] = version
        if may_not_exist:
            try:
                return self.get(**kwargs).state
            except Status.DoesNotExist:
                return None
        return self.get(**kwargs).state

    def version_for(self, id):
        return self.get(id=id).version

    def states_for(self, id_list):
        return [s.state for s in self.filter(id__in=id_list)]

    def prepare_for_pipeline(self, pipeline):
        cls_str = str(pipeline.__class__)
        cls_name = pipeline.__class__.__name__[:NAME_MAX_LENGTH]
        self.create(
            id=pipeline.id, state=states.READY, name=cls_str if len(cls_str) <= NAME_MAX_LENGTH else cls_name,
        )

    def fail(self, node, ex_data):
        action_res = self.transit(node.id, states.FAILED)

        if not action_res.result:
            return action_res

        Data.objects.write_node_data(node, ex_data)
        return action_res

    def raw_fail(self, node_id, ex_data):
        action_res = self.transit(node_id, states.FAILED)

        if not action_res.result:
            return action_res

        Data.objects.write_ex_data(node_id, ex_data)
        return action_res

    def finish(self, node, error_ignorable=False):
        action_res = self.transit(node.id, states.FINISHED)

        if not action_res.result:
            return action_res

        Data.objects.write_node_data(node)
        if error_ignorable:
            s = Status.objects.get(id=node.id)
            s.error_ignorable = True
            s.save()

        return action_res

    def skip(self, process, node):
        s = Status.objects.get(id=node.id)  # 一定要先取出来，不然 archive time 会被覆盖
        if RERUN_MAX_LIMIT != 0 and s.loop > RERUN_MAX_LIMIT:
            return ActionResult(
                result=False,
                message="rerun times exceed max limit: {limit}, can not skip".format(limit=RERUN_MAX_LIMIT),
            )

        action_res = self.transit(id=node.id, to_state=states.FINISHED, appoint=True)
        if not action_res.result:
            return action_res

        history = History.objects.record(s)
        LogEntry.objects.link_history(node_id=node.id, history_id=history.id)

        s.refresh_from_db()
        s.started_time = s.archived_time

        s.skip = True
        s.save()

        # 由于节点执行失败，数据可能尚未写入当前的 Data 对象，所以要在这里做一次写入操作
        node.skip()
        Data.objects.write_node_data(node)

        self.recover_from_block(process.root_pipeline.id, process.subprocess_stack)
        signals.node_skip_call.send(sender=Status, process=process, node=node)

        return action_res

    def retry(self, process, node, inputs):
        if RERUN_MAX_LIMIT != 0 and self.get(id=node.id).loop > RERUN_MAX_LIMIT:
            return ActionResult(
                result=False,
                message="rerun times exceed max limit: {limit}, can not retry".format(limit=RERUN_MAX_LIMIT),
            )

        action_res = self.transit(id=node.id, to_state=states.READY, appoint=True)
        if not action_res.result:
            return action_res

        # add retry times
        s = Status.objects.get(id=node.id)
        s.version = uniqid()
        history = History.objects.record(s)
        LogEntry.objects.link_history(node_id=node.id, history_id=history.id)
        s.retry += 1
        s.started_time = None
        s.archived_time = None
        s.save()

        # update inputs
        if inputs:
            new_data = DataObject(inputs=inputs, outputs={})
            node.data = new_data
            Data.objects.write_node_data(node)

        # mark
        node.next_exec_is_retry()

        self.recover_from_block(process.root_pipeline.id, process.subprocess_stack)
        signals.node_retry_ready.send(sender=Status, process=process, node=node)

        # because node may be updated
        process.save()

        return action_res

    def recover_from_block(self, root_pipeline_id, subprocess_stack):
        Status.objects.batch_transit(id_list=subprocess_stack, state=states.RUNNING, from_state=states.BLOCKED)
        Status.objects.transit(id=root_pipeline_id, to_state=states.READY, is_pipeline=True)

    @contextlib.contextmanager
    def lock(self, id):
        with transaction.atomic():
            self.select_for_update().get(id=id)
            yield


class Status(models.Model):
    id = models.CharField(_("节点 ID"), unique=True, primary_key=True, max_length=32)
    state = models.CharField(_("状态"), max_length=10)
    name = models.CharField(_("节点名称"), max_length=NAME_MAX_LENGTH, default="")
    retry = models.IntegerField(_("重试次数"), default=0)
    loop = models.IntegerField(_("循环次数"), default=1)
    skip = models.BooleanField(_("是否跳过"), default=False)
    error_ignorable = models.BooleanField(_("是否出错后自动忽略"), default=False)
    created_time = models.DateTimeField(_("创建时间"), auto_now_add=True, db_index=True)
    started_time = models.DateTimeField(_("开始时间"), null=True)
    archived_time = models.DateTimeField(_("归档时间"), null=True)
    version = models.CharField(_("版本"), max_length=32)
    state_refresh_at = models.DateTimeField(_("上次状态更新的时间"), null=True)

    objects = StatusManager()

    class Meta:
        ordering = ["-created_time"]

    def is_state_for_subproc(self):
        return self.name.endswith("SubProcess")


class DataManager(models.Manager):
    def write_node_data(self, node, ex_data=None):
        data, created = self.get_or_create(id=node.id)
        if hasattr(node, "data") and node.data:
            data.inputs = node.data.get_inputs()
            outputs = node.data.get_outputs()
            ex_data = outputs.pop("ex_data", ex_data)
            data.outputs = outputs
        data.ex_data = ex_data
        data.save()

    def write_ex_data(self, node_id, ex_data=None):
        data, created = self.get_or_create(id=node_id)
        data.ex_data = ex_data
        data.save()

    def forced_fail(self, node_id, ex_data=""):
        data, created = self.get_or_create(id=node_id)
        data.outputs = {
            "_forced_failed": True,
        }
        data.ex_data = ex_data
        data.save()


class Data(models.Model):
    id = models.CharField(_("节点 ID"), unique=True, primary_key=True, max_length=32)
    inputs = IOField(verbose_name=_("输入数据"), default=None)
    outputs = IOField(verbose_name=_("输出数据"), default=None)
    ex_data = IOField(verbose_name=_("异常数据"), default=None)

    objects = DataManager()


class HistoryData(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    inputs = IOField(verbose_name=_("输入数据"), default=None)
    outputs = IOField(verbose_name=_("输出数据"), default=None)
    ex_data = IOField(verbose_name=_("异常数据"), default=None)

    objects = DataManager()


class MultiCallbackData(models.Model):
    id = models.BigAutoField(_("自增ID"), primary_key=True)
    schedule_id = models.CharField(_("回调服务ID"), max_length=NAME_MAX_LENGTH)
    data = IOField(verbose_name=_("回调数据"))


DO_NOT_RECORD_WHEN_RERUN = frozenset({"<class 'pipeline.core.flow.activity.LoopServiceActivity'>"})


class HistoryManager(models.Manager):
    def record(self, status, is_rerunning=False):
        if is_rerunning and status.name in DO_NOT_RECORD_WHEN_RERUN:
            return None

        data = Data.objects.get(id=status.id)
        history_data = HistoryData.objects.create(inputs=data.inputs, outputs=data.outputs, ex_data=data.ex_data)
        return self.create(
            identifier=status.id,
            started_time=status.started_time,
            archived_time=status.archived_time,
            data=history_data,
            loop=status.loop,
            skip=status.skip,
        )

    def get_histories(self, identifier, loop=None):
        filters = {"identifier": identifier}
        if loop is not None:
            filters["loop"] = loop
        histories = self.filter(**filters).order_by("started_time")
        data = [
            {
                "history_id": item.id,
                "started_time": item.started_time,
                "archived_time": item.archived_time,
                "elapsed_time": calculate_elapsed_time(item.started_time, item.archived_time),
                "inputs": item.data.inputs,
                "outputs": item.data.outputs,
                "ex_data": item.data.ex_data,
                "loop": item.loop,
                "skip": item.skip,
            }
            for item in histories
        ]
        return data


class History(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    identifier = models.CharField(_("节点 id"), max_length=32, db_index=True)
    started_time = models.DateTimeField(_("开始时间"))
    archived_time = models.DateTimeField(_("结束时间"))
    loop = models.IntegerField(_("循环次数"), default=1)
    skip = models.BooleanField(_("是否跳过"), default=False)

    data = models.ForeignKey(HistoryData, null=True, on_delete=models.SET_NULL)

    objects = HistoryManager()


class ScheduleServiceManager(models.Manager):
    def set_schedule(self, activity_id, service_act, process_id, version, parent_data):
        wait_callback = service_act.service.interval is None
        multi_callback_enabled = service_act.service.multi_callback_enabled()
        schedule = self.create(
            id="{}{}".format(activity_id, version),
            activity_id=activity_id,
            service_act=service_act,
            process_id=process_id,
            wait_callback=wait_callback,
            multi_callback_enabled=multi_callback_enabled,
            version=version,
        )
        data_service.set_schedule_data(schedule.id, parent_data)

        if not wait_callback:
            count_down = service_act.service.interval.next()
            valve.send(
                signals,
                "schedule_ready",
                sender=ScheduleService,
                process_id=process_id,
                schedule_id=schedule.id,
                countdown=count_down,
            )

        return schedule

    def schedule_for(self, activity_id, version):
        return self.get(id="{}{}".format(activity_id, version))

    def delete_schedule(self, activity_id, version):
        return self.filter(activity_id=activity_id, version=version).delete()

    def update_celery_info(self, id, lock, celery_id, schedule_date, is_scheduling=False):
        return self.filter(id=id, celery_info_lock=lock).update(
            celery_info_lock=models.F("celery_info_lock") + 1,
            celery_id=celery_id,
            schedule_date=schedule_date,
            is_scheduling=is_scheduling,
        )


class ScheduleService(models.Model):
    SCHEDULE_ID_SPLIT_DIVISION = 32

    id = models.CharField(_("ID 节点ID+version"), max_length=NAME_MAX_LENGTH, unique=True, primary_key=True)
    activity_id = models.CharField(_("节点 ID"), max_length=32, db_index=True)
    process_id = models.CharField(_("Pipeline 进程 ID"), max_length=32)
    schedule_times = models.IntegerField(_("被调度次数"), default=0)
    wait_callback = models.BooleanField(_("是否是回调型调度"), default=False)
    multi_callback_enabled = models.BooleanField(_("是否支持多次回调"), default=False)
    callback_data = IOField(verbose_name=_("回调数据"), default=None)
    service_act = IOField(verbose_name=_("待调度服务"))
    is_finished = models.BooleanField(_("是否已完成"), default=False)
    version = models.CharField(_("Activity 的版本"), max_length=32, db_index=True)
    is_scheduling = models.BooleanField(_("是否正在被调度"), default=False, db_index=True)

    objects = ScheduleServiceManager()

    def set_next_schedule(self):
        if self.wait_callback:
            raise exceptions.InvalidOperationException("can't set next schedule on callback schedule.")
        count_down = self.service_act.service.interval.next()
        self.is_scheduling = False
        self.save()
        ScheduleCeleryTask.objects.unbind(self.id)

        valve.send(
            signals,
            "schedule_ready",
            sender=ScheduleService,
            process_id=self.process_id,
            schedule_id=self.id,
            countdown=count_down,
        )

    def destroy(self):
        schedule_id = self.id
        self.delete()
        data_service.delete_parent_data(schedule_id)
        ScheduleCeleryTask.objects.destroy(schedule_id)

    def finish(self):
        self.is_finished = True
        self.service_act = None
        self.is_scheduling = False
        self.save()
        ScheduleCeleryTask.objects.destroy(self.id)

    def callback(self, callback_data, process_id):
        if not self.wait_callback:
            raise exceptions.InvalidOperationException("can't callback a poll schedule.")

        if self.multi_callback_enabled:
            callback_data = MultiCallbackData.objects.create(schedule_id=self.id, data=callback_data)
            valve.send(
                signals,
                "schedule_ready",
                sender=ScheduleService,
                process_id=process_id,
                schedule_id=self.id,
                data_id=callback_data.id,
                countdown=0,
            )
        else:
            self.callback_data = callback_data
            self.save()
            valve.send(
                signals,
                "schedule_ready",
                sender=ScheduleService,
                process_id=process_id,
                schedule_id=self.id,
                countdown=0,
            )

    def is_one_time_callback(self):
        return self.wait_callback and not self.multi_callback_enabled


class SubProcessRelationshipManager(models.Manager):
    def add_relation(self, subprocess_id, process_id):
        return self.create(subprocess_id=subprocess_id, process_id=process_id)

    def delete_relation(self, subprocess_id, process_id):
        kwargs = {}
        if subprocess_id:
            kwargs["subprocess_id"] = subprocess_id
        if process_id:
            kwargs["process_id"] = process_id
        self.filter(**kwargs).delete()

    def get_relate_process(self, subprocess_id):
        qs = self.filter(subprocess_id=subprocess_id)
        proc_ids = [i.process_id for i in qs]
        return PipelineProcess.objects.filter(id__in=proc_ids)


class SubProcessRelationship(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    subprocess_id = models.CharField(_("子流程 ID"), max_length=32, db_index=True)
    process_id = models.CharField(_("对应的进程 ID"), max_length=32)

    objects = SubProcessRelationshipManager()


class ProcessCeleryTaskManager(models.Manager):
    def bind(self, process_id, celery_task_id):
        rel, created = self.get_or_create(process_id=process_id, defaults={"celery_task_id": celery_task_id})
        if not created:
            rel.celery_task_id = celery_task_id
            rel.save()

    def unbind(self, process_id):
        self.filter(process_id=process_id).update(celery_task_id="")

    def destroy(self, process_id):
        self.filter(process_id=process_id).delete()

    def start_task(self, process_id, task, kwargs, record_error=True):
        if record_error:
            with SendFailedCeleryTask.watch(
                name=task.name,
                kwargs=kwargs,
                type=SendFailedCeleryTask.TASK_TYPE_PROCESS,
                extra_kwargs={"process_id": process_id},
            ):
                task_id = task.apply_async(**kwargs)
        else:
            task_id = task.apply_async(**kwargs)
        self.bind(process_id, task_id)

    def revoke(self, process_id, kill=False):
        task = self.get(process_id=process_id)
        kwargs = {} if not kill else {"signal": "SIGKILL"}
        revoke(task.celery_task_id, terminate=True, **kwargs)
        self.destroy(process_id)


class ProcessCeleryTask(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    process_id = models.CharField(_("pipeline 进程 ID"), max_length=32, unique=True, db_index=True)
    celery_task_id = models.CharField(_("celery 任务 ID"), max_length=40, default="")

    objects = ProcessCeleryTaskManager()


class ScheduleCeleryTaskManager(models.Manager):
    def bind(self, schedule_id, celery_task_id):
        rel, created = self.get_or_create(schedule_id=schedule_id, defaults={"celery_task_id": celery_task_id})
        if not created:
            rel.celery_task_id = celery_task_id
            rel.save()

    def unbind(self, schedule_id):
        self.filter(schedule_id=schedule_id).update(celery_task_id="")

    def destroy(self, schedule_id):
        self.filter(schedule_id=schedule_id).delete()

    def start_task(self, schedule_id, task, kwargs, record_error=True):
        if record_error:
            with SendFailedCeleryTask.watch(
                name=task.name,
                kwargs=kwargs,
                type=SendFailedCeleryTask.TASK_TYPE_SCHEDULE,
                extra_kwargs={"schedule_id": schedule_id},
            ):
                task_id = task.apply_async(**kwargs)
        else:
            task_id = task.apply_async(**kwargs)
        self.bind(schedule_id, task_id)


class ScheduleCeleryTask(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    schedule_id = models.CharField(_("schedule ID"), max_length=NAME_MAX_LENGTH, unique=True, db_index=True)
    celery_task_id = models.CharField(_("celery 任务 ID"), max_length=40, default="")

    objects = ScheduleCeleryTaskManager()


class NodeCeleryTaskManager(models.Manager):
    def bind(self, node_id, celery_task_id):
        rel, created = self.update_or_create(node_id=node_id, defaults={"celery_task_id": celery_task_id})
        if not created:
            rel.celery_task_id = celery_task_id
            rel.save()

    def unbind(self, node_id):
        self.filter(node_id=node_id).update(celery_task_id="")

    def destroy(self, node_id):
        self.filter(node_id=node_id).delete()

    def start_task(self, node_id, task, kwargs, record_error=True):
        if record_error:
            with SendFailedCeleryTask.watch(
                name=task.name,
                kwargs=kwargs,
                type=SendFailedCeleryTask.TASK_TYPE_NODE,
                extra_kwargs={"node_id": node_id},
            ):
                task_id = task.apply_async(**kwargs)
        else:
            task_id = task.apply_async(**kwargs)
        self.bind(node_id, task_id)

    def revoke(self, node_id):
        task = self.get(node_id=node_id)
        revoke(task.celery_task_id)
        self.destroy(node_id)


class NodeCeleryTask(models.Model):
    id = models.BigAutoField(_("ID"), primary_key=True)
    node_id = models.CharField(_("节点 ID"), max_length=32, unique=True, db_index=True)
    celery_task_id = models.CharField(_("celery 任务 ID"), max_length=40, default="")

    objects = NodeCeleryTaskManager()


class SendFailedCeleryTaskManager(models.Manager):
    def record(self, name, kwargs, type, extra_kwargs, exec_trace):
        save_extra_kwargs = extra_kwargs
        save_kwargs = kwargs

        if not isinstance(save_extra_kwargs, str):
            save_extra_kwargs = json.dumps(save_extra_kwargs)

        if not isinstance(save_kwargs, str):
            save_kwargs = json.dumps(save_kwargs)

        return self.create(
            name=name, kwargs=save_kwargs, type=type, extra_kwargs=save_extra_kwargs, exec_trace=exec_trace,
        )

    def resend(self, id):
        self.get(id=id).resend()


class SendFailedCeleryTask(models.Model):

    TASK_TYPE_EMPTY = 0
    TASK_TYPE_PROCESS = 1
    TASK_TYPE_NODE = 2
    TASK_TYPE_SCHEDULE = 3

    TASK_TYPE_CHOICES = (
        (TASK_TYPE_EMPTY, "empty"),
        (TASK_TYPE_PROCESS, "process"),
        (TASK_TYPE_NODE, "node"),
        (TASK_TYPE_SCHEDULE, "schedule"),
    )

    id = models.BigAutoField(_("ID"), primary_key=True)
    name = models.CharField(_("任务名"), max_length=1024)
    kwargs = models.TextField(_("任务参数"))
    type = models.IntegerField(_("任务类型"), choices=TASK_TYPE_CHOICES)
    extra_kwargs = models.TextField(_("额外参数"))
    exec_trace = models.TextField(_("错误信息"))
    created_at = models.DateTimeField(_("创建时间"), auto_now_add=True)

    objects = SendFailedCeleryTaskManager()

    @property
    def kwargs_dict(self):
        return json.loads(self.kwargs)

    @property
    def extra_kwargs_dict(self):
        return json.loads(self.extra_kwargs)

    def resend(self):
        try:
            task = current_app.tasks[self.name]

            if self.type == self.TASK_TYPE_EMPTY:
                task.apply_async(**self.kwargs_dict)
            elif self.type == self.TASK_TYPE_PROCESS:
                ProcessCeleryTask.objects.start_task(
                    process_id=self.extra_kwargs_dict["process_id"],
                    task=task,
                    kwargs=self.kwargs_dict,
                    record_error=False,
                )
            elif self.type == self.TASK_TYPE_NODE:
                NodeCeleryTask.objects.start_task(
                    node_id=self.extra_kwargs_dict["node_id"], task=task, kwargs=self.kwargs_dict, record_error=False,
                )
            elif self.type == self.TASK_TYPE_SCHEDULE:
                ScheduleCeleryTask.objects.start_task(
                    schedule_id=self.extra_kwargs_dict["schedule_id"],
                    task=task,
                    kwargs=self.kwargs_dict,
                    record_error=False,
                )
            else:
                raise TypeError("unsupport type: {}.".format(self.type))
        except Exception as e:
            logger.exception("fail task send replay error.")
            raise e
        else:
            self.delete()

    @classmethod
    @contextlib.contextmanager
    def watch(cls, name, kwargs, type, extra_kwargs):
        try:
            yield
        except Exception:
            logger.exception("celery task({}) watcher catch error.".format(name))
            cls.objects.record(
                name=name, kwargs=kwargs, type=type, extra_kwargs=extra_kwargs, exec_trace=traceback.format_exc(),
            )
            # raise specific exception to indicate that send fail task have been catched
            raise exceptions.CeleryFailedTaskCatchException(name)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import models
from django.utils.translation import ugettext_lazy as _


class ComponentInTemplate(models.Model):
    component_code = models.CharField(_("组件编码"), max_length=255)
    template_id = models.CharField(_("模板ID"), max_length=32)
    node_id = models.CharField(_("节点ID"), max_length=32)
    is_sub = models.BooleanField(_("是否子流程引用"), default=False)
    subprocess_stack = models.TextField(_("子流程堆栈"), default="[]", help_text=_("JSON 格式的列表"))
    version = models.CharField(_("插件版本"), max_length=255, default="legacy")

    class Meta:
        verbose_name = _("Pipeline标准插件被引用数据")
        verbose_name_plural = _("Pipeline标准插件被引用数据")

    def __unicode__(self):
        return "{}_{}".format(self.component_code, self.template_id)


class ComponentExecuteData(models.Model):
    component_code = models.CharField(_("组件编码"), max_length=255, db_index=True)
    instance_id = models.CharField(_("实例ID"), max_length=32, db_index=True)
    node_id = models.CharField(_("节点ID"), max_length=32)
    is_sub = models.BooleanField(_("是否子流程引用"), default=False)
    subprocess_stack = models.TextField(_("子流程堆栈"), default="[]", help_text=_("JSON 格式的列表"))
    started_time = models.DateTimeField(_("标准插件执行开始时间"))
    archived_time = models.DateTimeField(_("标准插件执行结束时间"), null=True, blank=True)
    elapsed_time = models.IntegerField(_("标准插件执行耗时(s)"), null=True, blank=True)
    status = models.BooleanField(_("是否执行成功"), default=False)
    is_skip = models.BooleanField(_("是否跳过"), default=False)
    is_retry = models.BooleanField(_("是否重试记录"), default=False)
    version = models.CharField(_("插件版本"), max_length=255, default="legacy")

    class Meta:
        verbose_name = _("Pipeline标准插件执行数据")
        verbose_name_plural = _("Pipeline标准插件执行数据")
        ordering = ["-id"]

    def __unicode__(self):
        return "{}_{}".format(self.component_code, self.instance_id)


class TemplateInPipeline(models.Model):
    template_id = models.CharField(_("模板ID"), max_length=255, db_index=True)
    atom_total = models.IntegerField(_("标准插件总数"))
    subprocess_total = models.IntegerField(_("子流程总数"))
    gateways_total = models.IntegerField(_("网关总数"))

    class Meta:
        verbose_name = _("Pipeline模板引用数据")
        verbose_name_plural = _("Pipeline模板引用数据")

    def __unicode__(self):
        return "{}_{}_{}_{}".format(self.template_id, self.atom_total, self.subprocess_total, self.gateways_total)


class InstanceInPipeline(models.Model):
    instance_id = models.CharField(_("实例ID"), max_length=255, db_index=True)
    atom_total = models.IntegerField(_("标准插件总数"))
    subprocess_total = models.IntegerField(_("子流程总数"))
    gateways_total = models.IntegerField(_("网关总数"))

    class Meta:
        verbose_name = _("Pipeline实例引用数据")
        verbose_name_plural = _("Pipeline实例引用数据")

    def __unicode__(self):
        return "{}_{}_{}_{}".format(self.instance_id, self.atom_total, self.subprocess_total, self.gateways_total)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from abc import ABCMeta, abstractmethod
from copy import deepcopy

from django.utils.translation import ugettext_lazy as _

from pipeline.core.flow.activity.base import Activity
from pipeline.core.flow.io import BooleanItemSchema, InputItem, IntItemSchema, OutputItem
from pipeline.utils.utils import convert_bytes_to_str


class Service(object, metaclass=ABCMeta):
    schedule_result_attr = "__schedule_finish__"
    schedule_determine_attr = "__need_schedule__"
    multi_callback_determine_attr = "__multi_callback_enabled__"
    InputItem = InputItem
    OutputItem = OutputItem
    interval = None
    default_outputs = [
        OutputItem(
            name=_("执行结果"),
            key="_result",
            type="boolean",
            schema=BooleanItemSchema(description=_("执行结果的布尔值，True or False")),
        ),
        OutputItem(name=_("循环次数"), key="_loop", type="int", schema=IntItemSchema(description=_("循环执行次数"))),
    ]

    def __init__(self, name=None):
        self.name = name
        self.interval = deepcopy(self.interval)
        self._runtime_attrs = {}

    def __getattr__(self, name):
        if name not in self.__dict__.get("_runtime_attrs", {}):
            raise AttributeError()

        return self._runtime_attrs[name]

    def __getstate__(self):
        if "logger" in self.__dict__:
            del self.__dict__["logger"]
        # compatible with old version pickle obj
        if "_runtime_attrs" in self.__dict__:
            if "logger" in self._runtime_attrs:
                del self._runtime_attrs["logger"]

        return self.__dict__

    @abstractmethod
    def execute(self, data, parent_data):
        # get params from data
        pass

    def outputs_format(self):
        return []

    def inputs_format(self):
        return []

    def inputs(self):
        return self.inputs_format()

    def outputs(self):
        custom_format = self.outputs_format()
        assert isinstance(custom_format, list)
        custom_format += self.default_outputs
        return custom_format

    def need_schedule(self):
        return getattr(self, Service.schedule_determine_attr, False)

    def schedule(self, data, parent_data, callback_data=None):
        return True

    def finish_schedule(self):
        setattr(self, self.schedule_result_attr, True)

    def is_schedule_finished(self):
        return getattr(self, self.schedule_result_attr, False)

    def multi_callback_enabled(self):
        return getattr(self, self.multi_callback_determine_attr, False)

    def clean_status(self):
        setattr(self, self.schedule_result_attr, False)

    def setup_runtime_attrs(self, **kwargs):
        # compatible with old version pickle obj
        if "_runtime_attrs" not in self.__dict__:
            self._runtime_attrs = {}
        self._runtime_attrs.update(**kwargs)


class ServiceActivity(Activity):
    result_bit = "_result"
    loop = "_loop"
    ON_RETRY = "_on_retry"

    def __init__(
        self,
        id,
        service,
        name=None,
        data=None,
        error_ignorable=False,
        failure_handler=None,
        skippable=True,
        retryable=True,
        timeout=None,
    ):
        super(ServiceActivity, self).__init__(id, name, data, failure_handler)
        self.service = service
        self.error_ignorable = error_ignorable
        self.skippable = skippable
        self.retryable = retryable
        self.timeout = timeout

        if data:
            self._prepared_inputs = self.data.inputs_copy()
            self._prepared_outputs = self.data.outputs_copy()

    def __setstate__(self, state):

        for attr, obj in list(state.items()):
            # py2 pickle dumps data compatible
            if isinstance(attr, bytes):
                attr = attr.decode("utf-8")
                obj = convert_bytes_to_str(obj)

            setattr(self, attr, obj)

        if "timeout" not in state:
            self.timeout = None

    def execute_pre_process(self, parent_data):
        # return True if the plugin does not complete execute_pre_process function
        if not (hasattr(self.service, "execute_pre_process") and callable(self.service.execute_pre_process)):
            return True

        result = self.service.execute_pre_process(self.data, parent_data)

        # set result
        self.set_result_bit(result)

        if self.error_ignorable:
            return True
        return result

    def execute(self, parent_data):
        self.setup_logger()
        result = self.service.execute(self.data, parent_data)

        # set result
        self.set_result_bit(result)

        if self.error_ignorable:
            return True
        return result

    def set_result_bit(self, result):
        if result is False:
            self.data.set_outputs(self.result_bit, False)
        else:
            self.data.set_outputs(self.result_bit, True)

    def get_result_bit(self):
        return self.data.get_one_of_outputs(self.result_bit, False)

    def skip(self):
        self.set_result_bit(True)
        return True

    def ignore_error(self):
        self.set_result_bit(False)
        return True

    def clear_outputs(self):
        self.data.reset_outputs({})

    def need_schedule(self):
        return self.service.need_schedule()

    def schedule(self, parent_data, callback_data=None):
        self.setup_logger()
        result = self.service.schedule(self.data, parent_data, callback_data)
        self.set_result_bit(result)

        if result is False:
            if self.error_ignorable:
                self.service.finish_schedule()
                return True

        return result

    def is_schedule_done(self):
        return self.service.is_schedule_finished()

    def finish_schedule(self):
        self.service.finish_schedule()

    def shell(self):
        shell = ServiceActivity(
            id=self.id,
            service=self.service,
            name=self.name,
            data=self.data,
            error_ignorable=self.error_ignorable,
            timeout=self.timeout,
        )
        return shell

    def schedule_fail(self):
        return

    def schedule_success(self):
        return

    def prepare_rerun_data(self):
        self.data.override_inputs(deepcopy(self._prepared_inputs))
        self.data.override_outputs(deepcopy(self._prepared_outputs))

    def setup_runtime_attrs(self, **kwargs):
        self.service.setup_runtime_attrs(**kwargs)

    def setup_logger(self):
        self.service.setup_runtime_attrs(logger=self.logger)


class AbstractIntervalGenerator(object, metaclass=ABCMeta):
    def __init__(self):
        self.count = 0

    def next(self):
        self.count += 1


class DefaultIntervalGenerator(AbstractIntervalGenerator):
    def next(self):
        super(DefaultIntervalGenerator, self).next()
        return self.count ** 2


class SquareIntervalGenerator(AbstractIntervalGenerator):
    def next(self):
        super(SquareIntervalGenerator, self).next()
        return self.count ** 2


class NullIntervalGenerator(AbstractIntervalGenerator):
    pass


class LinearIntervalGenerator(AbstractIntervalGenerator):
    pass


class StaticIntervalGenerator(AbstractIntervalGenerator):
    def __init__(self, interval):
        super(StaticIntervalGenerator, self).__init__()
        self.interval = interval

    def next(self):
        super(StaticIntervalGenerator, self).next()
        return self.interval



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import datetime
import logging
import traceback

import pytz
from celery import task
from django.utils import timezone
from django.utils.module_loading import import_string
from bamboo_engine import api as bamboo_engine_api

from pipeline.contrib.periodic_task import signals
from pipeline.contrib.periodic_task.models import PeriodicTask, PeriodicTaskHistory
from pipeline.engine.models import FunctionSwitch
from pipeline.models import PipelineInstance
from pipeline.parser.context import get_pipeline_context
from pipeline.eri.runtime import BambooDjangoRuntime

logger = logging.getLogger("celery")


@task(ignore_result=True)
def periodic_task_start(*args, **kwargs):
    try:
        periodic_task = PeriodicTask.objects.get(id=kwargs["period_task_id"])
    except PeriodicTask.DoesNotExist:
        # task has been deleted
        return

    if FunctionSwitch.objects.is_frozen():
        PeriodicTaskHistory.objects.record_schedule(
            periodic_task=periodic_task,
            pipeline_instance=None,
            ex_data="engine is frozen, can not start task",
            start_success=False,
        )
        return

    try:
        tz = periodic_task.celery_task.crontab.timezone
        now = datetime.datetime.now(tz=pytz.utc).astimezone(tz)
        instance, _ = PipelineInstance.objects.create_instance(
            template=periodic_task.template,
            exec_data=periodic_task.execution_data,
            spread=kwargs.get("spread", True),
            name="{}_{}".format(periodic_task.name[:113], now.strftime("%Y%m%d%H%M%S")),
            creator=periodic_task.creator,
            description="periodic task instance",
        )

        signals.pre_periodic_task_start.send(
            sender=PeriodicTask, periodic_task=periodic_task, pipeline_instance=instance
        )

        result = instance.start(
            periodic_task.creator, check_workers=False, priority=periodic_task.priority, queue=periodic_task.queue,
        )
    except Exception:
        et = traceback.format_exc()
        logger.error(et)
        PeriodicTaskHistory.objects.record_schedule(
            periodic_task=periodic_task, pipeline_instance=None, ex_data=et, start_success=False,
        )
        return

    if not result.result:
        PeriodicTaskHistory.objects.record_schedule(
            periodic_task=periodic_task, pipeline_instance=None, ex_data=result.message, start_success=False,
        )
        return

    periodic_task.total_run_count += 1
    periodic_task.last_run_at = timezone.now()
    periodic_task.save()
    signals.post_periodic_task_start.send(sender=PeriodicTask, periodic_task=periodic_task, pipeline_instance=instance)

    PeriodicTaskHistory.objects.record_schedule(periodic_task=periodic_task, pipeline_instance=instance, ex_data="")


@task(ignore_result=True)
def bamboo_engine_periodic_task_start(*args, **kwargs):
    try:
        periodic_task = PeriodicTask.objects.get(id=kwargs["period_task_id"])
    except PeriodicTask.DoesNotExist:
        # task has been deleted
        return

    try:
        tz = periodic_task.celery_task.crontab.timezone
        now = datetime.datetime.now(tz=pytz.utc).astimezone(tz)
        instance, _ = PipelineInstance.objects.create_instance(
            template=periodic_task.template,
            exec_data=periodic_task.execution_data,
            spread=kwargs.get("spread", True),
            name="{}_{}".format(periodic_task.name[:113], now.strftime("%Y%m%d%H%M%S")),
            creator=periodic_task.creator,
            description="periodic task instance",
        )

        signals.pre_periodic_task_start.send(
            sender=PeriodicTask, periodic_task=periodic_task, pipeline_instance=instance
        )

        # convert web pipeline to pipeline
        pipeline_formator = import_string(periodic_task.extra_info["pipeline_formator"])
        pipeline = pipeline_formator(instance.execution_data)

        # run pipeline
        instance.calculate_tree_info()
        PipelineInstance.objects.filter(instance_id=instance.instance_id).update(
            tree_info_id=instance.tree_info.id,
            start_time=timezone.now(),
            is_started=True,
            executor=periodic_task.creator,
        )
        result = bamboo_engine_api.run_pipeline(
            runtime=BambooDjangoRuntime(),
            pipeline=pipeline,
            root_pipeline_data=get_pipeline_context(
                instance, obj_type="instance", data_type="data", username=periodic_task.creator
            ),
            queue=periodic_task.queue,
        )
    except Exception:
        et = traceback.format_exc()
        logger.error(et)
        PeriodicTaskHistory.objects.record_schedule(
            periodic_task=periodic_task, pipeline_instance=None, ex_data=et, start_success=False,
        )
        return

    if not result.result:
        PipelineInstance.objects.filter(id=instance.instance_id).update(
            start_time=None, is_started=False, executor="",
        )
        PeriodicTaskHistory.objects.record_schedule(
            periodic_task=periodic_task, pipeline_instance=None, ex_data=result.message, start_success=False,
        )
        return

    periodic_task.total_run_count += 1
    periodic_task.last_run_at = timezone.now()
    periodic_task.save()
    signals.post_periodic_task_start.send(sender=PeriodicTask, periodic_task=periodic_task, pipeline_instance=instance)

    PeriodicTaskHistory.objects.record_schedule(periodic_task=periodic_task, pipeline_instance=instance, ex_data="")



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.utils.translation import ugettext_lazy as _

from pipeline.exceptions import ConnectionValidateError
from pipeline.utils.graph import Graph
from pipeline.validators.rules import NODE_RULES
from pipeline.validators.utils import get_nodes_dict
from pipeline.core.constants import PE


def validate_graph_connection(data):
    """
    节点连接合法性校验
    """
    nodes = get_nodes_dict(data)

    result = {"result": True, "message": {}, "failed_nodes": []}

    for i in nodes:
        node_type = nodes[i][PE.type]
        rule = NODE_RULES[node_type]
        message = ""
        for j in nodes[i][PE.target]:
            if nodes[j][PE.type] not in rule["allowed_out"]:
                message += _("不能连接%s类型节点\n") % nodes[i][PE.type]
            if rule["min_in"] > len(nodes[i][PE.source]) or len(nodes[i][PE.source]) > rule["max_in"]:
                message += _("节点的入度最大为%s，最小为%s\n") % (rule["max_in"], rule["min_in"])
            if rule["min_out"] > len(nodes[i][PE.target]) or len(nodes[i][PE.target]) > rule["max_out"]:
                message += _("节点的出度最大为%s，最小为%s\n") % (rule["max_out"], rule["min_out"])
        if message:
            result["failed_nodes"].append(i)
            result["message"][i] = message

        if result["failed_nodes"]:
            raise ConnectionValidateError(failed_nodes=result["failed_nodes"], detail=result["message"])


def validate_graph_without_circle(data):
    """
    validate if a graph has not cycle

    return {
        "result": False,
        "message": "error message",
        "error_data": ["node1_id", "node2_id", "node1_id"]
    }
    """

    nodes = [data[PE.start_event][PE.id], data[PE.end_event][PE.id]]
    nodes += list(data[PE.gateways].keys()) + list(data[PE.activities].keys())
    flows = [[flow[PE.source], flow[PE.target]] for _, flow in list(data[PE.flows].items())]
    cycle = Graph(nodes, flows).get_cycle()
    if cycle:
        return {"result": False, "message": "pipeline graph has circle", "error_data": cycle}
    return {"result": True, "data": []}



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import ast
from typing import List

from mako import parsetree
from mako.exceptions import MakoException
from mako.lexer import Lexer

from .code_extract import MakoNodeCodeExtractor
from .exceptions import ForbiddenMakoTemplateException


def parse_template_nodes(
    nodes: List[parsetree.Node], node_visitor: ast.NodeVisitor, code_extractor: MakoNodeCodeExtractor,
):
    """
    解析mako模板节点，逐个节点解析抽象语法树并检查安全性
    :param nodes: mako模板节点列表
    :param node_visitor: 节点访问类，用于遍历AST节点
    :param code_extractor: Mako 词法节点处理器，用于提取 python 代码
    """
    for node in nodes:
        code = code_extractor.extract(node)
        if code is None:
            continue

        ast_node = ast.parse(code, "<unknown>", "exec")
        node_visitor.visit(ast_node)
        if hasattr(node, "nodes"):
            parse_template_nodes(node.nodes, node_visitor)


def check_mako_template_safety(text: str, node_visitor: ast.NodeVisitor, code_extractor: MakoNodeCodeExtractor) -> bool:
    """
    检查mako模板是否安全，若不安全直接抛出异常，安全则返回True
    :param text: mako模板内容
    :param node_visitor: 节点访问器，用于遍历AST节点
    """
    try:
        lexer_template = Lexer(text).parse()
    except MakoException as mako_error:
        raise ForbiddenMakoTemplateException("非mako模板，解析失败, {err_msg}".format(err_msg=mako_error.__class__.__name__))
    parse_template_nodes(lexer_template.nodes, node_visitor, code_extractor)
    return True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-12-01 07:06
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("statistics", "0012_auto_20201123_1552"),
    ]

    operations = [
        migrations.AlterField(
            model_name="componentexecutedata",
            name="component_code",
            field=models.CharField(db_index=True, max_length=255, verbose_name="组件编码"),
        ),
        migrations.AlterField(
            model_name="componentexecutedata",
            name="instance_id",
            field=models.CharField(db_index=True, max_length=32, verbose_name="实例ID"),
        ),
    ]
